# encoding: utf-8
"""
match.py

Created by Thomas Mangin on 2015-06-22.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from exabgp.configuration.core import Section

from exabgp.configuration.flow.parser import source
from exabgp.configuration.flow.parser import destination
from exabgp.configuration.flow.parser import any_port
from exabgp.configuration.flow.parser import source_port
from exabgp.configuration.flow.parser import destination_port
from exabgp.configuration.flow.parser import tcp_flags
from exabgp.configuration.flow.parser import protocol
from exabgp.configuration.flow.parser import next_header
from exabgp.configuration.flow.parser import fragment
from exabgp.configuration.flow.parser import packet_length
from exabgp.configuration.flow.parser import icmp_code
from exabgp.configuration.flow.parser import icmp_type
from exabgp.configuration.flow.parser import dscp
from exabgp.configuration.flow.parser import traffic_class
from exabgp.configuration.flow.parser import flow_label


class ParseFlowMatch (Section):
	definition = [
		'source 10.0.0.0/24',
		'source ::1/128/0',
		'destination 10.0.1.0/24',
		'port 25',
		'source-port >1024',
		'destination-port [ =80 =3128 >8080&<8088 ]',
		'packet-length [ >200&<300 >400&<500 ]',
		'tcp-flags [ 0x20+0x8+0x1 #name-here ]  # to check',
		'(ipv4 only) protocol [ udp tcp ]',
		'(ipv4 only) fragment [ not-a-fragment dont-fragment is-fragment first-fragment last-fragment ]',
		'(ipv6 only) next-header [ udp tcp ]',
		'(ipv6 only) flow-label >100&<2000',
		'(ipv6 only) icmp-type 35  # to check',
		'(ipv6 only) icmp-code 55  # to check',
	]

	syntax = \
		'match {\n' \
		'  %s;\n' \
		'}' % ';\n  '.join(definition)

	known = {
		'source':           source,
		'source-ipv4':      source,
		'source-ipv6':      source,
		'destination':      destination,
		'destination-ipv4': destination,
		'destination-ipv6': destination,
		'protocol':         protocol,
		'next-header':      next_header,
		'port':             any_port,
		'destination-port': destination_port,
		'source-port':      source_port,
		'icmp-type':        icmp_type,
		'icmp-code':        icmp_code,
		'tcp-flags':        tcp_flags,
		'packet-length':    packet_length,
		'dscp':             dscp,
		'traffic-class':    traffic_class,
		'fragment':         fragment,
		'flow-label':       flow_label,
	}

	# 'source-ipv4','destination-ipv4',

	action = {
		'source':           'nlri-add',
		'source-ipv4':      'nlri-add',
		'source-ipv6':      'nlri-add',
		'destination':      'nlri-add',
		'destination-ipv4': 'nlri-add',
		'destination-ipv6': 'nlri-add',
		'port':             'nlri-add',
		'source-port':      'nlri-add',
		'destination-port': 'nlri-add',
		'protocol':         'nlri-add',
		'packet-length':    'nlri-add',
		'tcp-flags':        'nlri-add',
		'next-header':      'nlri-add',
		'fragment':         'nlri-add',
		'icmp-code':        'nlri-add',
		'icmp-type':        'nlri-add',
		'packet-length':    'nlri-add',
		'dscp':             'nlri-add',
		'traffic-class':    'nlri-add',
		'flow-label':       'nlri-add',
	}

	name = 'flow/match'

	def __init__ (self, tokeniser, scope, error, logger):
		Section.__init__(self,tokeniser,scope,error,logger)

	def clear (self):
		pass

	def pre (self):
		return True

	def post (self):
		return True



# encoding: utf-8
"""
Copyright (c) 2016 Job Snijders <job@ntt.net>
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from exabgp.util import ordinal
from exabgp.bgp.message.update.attribute import Attribute
from exabgp.bgp.message.update.attribute.community.initial import Communities
from exabgp.bgp.message.update.attribute.community.large import LargeCommunity

from exabgp.bgp.message.notification import Notify


@Attribute.register()
class LargeCommunities (Communities):
	ID = Attribute.CODE.LARGE_COMMUNITY

	@staticmethod
	def unpack (data, negotiated):
		large_communities = LargeCommunities()
		while data:
			if data and len(data) < 12:
				raise Notify(3,1,'could not decode large community %s' % str([hex(ordinal(_)) for _ in data]))
			lc = LargeCommunity.unpack(data[:12],negotiated)
			data = data[12:]
			if lc in large_communities.communities:
				continue
			large_communities.add(lc)
		return large_communities



# encoding: utf-8
"""
capability.py

Created by Thomas Mangin on 2015-06-04.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from exabgp.bgp.message.open.capability.graceful import Graceful

from exabgp.configuration.core import Section
from exabgp.configuration.parser import boolean
from exabgp.configuration.parser import string


def addpath (tokeniser):
	if not tokeniser.tokens:
		raise ValueError('add-path must be one of send, receive, send/receive, disable')

	ap = string(tokeniser).lower()

	match = {
		'disable':      0,
		'disabled':     0,
		'receive':      1,
		'send':         2,
		'send/receive': 3,
	}

	if ap in match:
		return match[ap]

	if ap == 'receive/send':  # was allowed with the previous parser
		raise ValueError('the option is send/receive')

	raise ValueError('"%s" is an invalid add-path, options are: send, receive, send/receive' % ap)


def gracefulrestart (tokeniser, default):
	if len(tokeniser.tokens) == 1:
		return default

	state = string(tokeniser)

	if state in ('disable','disabled'):
		return False

	try:
		grace = int(state)
	except ValueError:
		raise ValueError('"%s" is an invalid graceful-restart time' % state)

	if grace < 0:
		raise ValueError('graceful-restart can not be negative')
	if grace > Graceful.MAX:
		raise ValueError('graceful-restart must be smaller or equal to %d' % Graceful.MAX)

	return grace


class ParseCapability (Section):
	TTL_SECURITY = 255

	syntax = \
		'capability {\n' \
		'   add-path disable|send|receive|send/receive;\n' \
		'   asn4 enable|disable;\n' \
		'   graceful-restart <time in second>;\n' \
		'   multi-session enable|disable;\n' \
		'   operational enable|disable;\n' \
		'   refresh enable|disable;\n' \
		'   extended-message enable|disable;\n' \
		'}\n'

	known = {
		'add-path':         addpath,
		'asn4':             boolean,
		'graceful-restart': gracefulrestart,
		'multi-session':    boolean,
		'operational':      boolean,
		'route-refresh':    boolean,
		'aigp':             boolean,
		'extended-message': boolean,
	}

	action = {
		'add-path':         'set-command',
		'asn4':             'set-command',
		'graceful-restart': 'set-command',
		'multi-session':    'set-command',
		'operational':      'set-command',
		'route-refresh':    'set-command',
		'aigp':             'set-command',
		'extended-message': 'set-command',
	}

	default = {
		'asn4':             True,
		'graceful-restart': 0,
		'multi-session':    True,
		'operational':      True,
		'route-refresh':    True,
		'aigp':             True,
		'extended-message': True,
	}

	name = 'capability'

	def __init__ (self, tokeniser, scope, error, logger):
		Section.__init__(self,tokeniser,scope,error,logger)

	def pre (self):
		return True

	def post (self):
		return True

	def clear (self):
		pass



# encoding: utf-8
"""
message.py

Created by Thomas Mangin on 2015-03-31.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

import socket
from struct import unpack
from collections import namedtuple

from exabgp.netlink.attributes import Attributes
from exabgp.netlink.netlink import NetLink


class Message (object):
	# to be defined by the subclasses
	format = namedtuple('Parent', 'to be subclassed')

	DEFAULT_FLAGS = NetLink.Flags.NLM_F_REQUEST | NetLink.Flags.NLM_F_DUMP

	# to be defined by the subclasses
	class Header (object):
		PACK = ''
		LEN = 0

	@classmethod
	def decode (cls, data):
		extracted = list(unpack(cls.Header.PACK,data[:cls.Header.LEN]))
		attributes = Attributes.decode(data[cls.Header.LEN:])
		extracted.append(dict(attributes))
		return cls.format(*extracted)

	@classmethod
	def extract (cls, format_type, control_flags=DEFAULT_FLAGS, family=socket.AF_UNSPEC, attributes=None):
		for data in NetLink.send(format_type,control_flags,family,attributes):
			yield cls.decode(data)



# encoding: utf-8
"""
generic/parser.py

Created by Thomas Mangin on 2015-06-04.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

import socket

from exabgp.bgp.message.open.asn import ASN
from exabgp.protocol.ip import IP
from exabgp.protocol.ip import IPRange


def string (tokeniser):
	return tokeniser()


def boolean (tokeniser, default):
	status = tokeniser().lower()
	if not status:
		return default
	if status in ('true','enable','enabled'):
		return True
	if status in ('false','disable','disabled'):
		return False
	raise ValueError('invalid value (%s) for a boolean' % status)


def auto_boolean (tokeniser, default):
	status = tokeniser().lower()
	if not status:
		return default
	if status in ('true','enable','enabled'):
		return True
	if status in ('false','disable','disabled'):
		return False
	if status in ('auto',):
		return None
	raise ValueError('invalid value (%s) for a boolean' % status)


def port (tokeniser):
	if not tokeniser.tokens:
		raise ValueError('a port number is required')

	value = tokeniser()
	try:
		return int(value)
	except ValueError:
		raise ValueError('"%s" is an invalid port' % value)
	if value < 0:
		raise ValueError('the port must be positive')
	if value >= pow(2,16):
		raise ValueError('the port must be smaller than %d' % pow(2,16))
	return value


def auto_asn (tokeniser, value=None):
	if value is None:
		if not tokeniser.tokens:
			raise ValueError("an asn or 'auto' is required")

	if tokeniser.peek() == 'auto':
		tokeniser()
		return None

	return asn(tokeniser)


def asn (tokeniser, value=None):
	if value is None:
		if not tokeniser.tokens:
			raise ValueError('an asn is required')

	value = tokeniser()
	try:
		if value.count('.'):
			high,low = value.split('.',1)
			as_number = (int(high) << 16) + int(low)
		else:
			as_number = int(value)
		return ASN(as_number)
	except ValueError:
		raise ValueError('"%s" is an invalid ASN' % value)


def peer_ip (tokeniser):
	if not tokeniser.tokens:
		raise ValueError('an ip address is required')

	value = tokeniser()
	if '/' in value:
		value,mask = value.split('/',1)
	else:
		# XXX: This only works as no port are allowed, improve
		mask = '128' if ':' in value else '32'

	try:
		return IPRange.create(value,mask)
	except (IndexError,ValueError,socket.error):
		raise ValueError('"%s" is an invalid IP address' % value)


def ip (tokeniser):
	if not tokeniser.tokens:
		raise ValueError('an ip address is required')

	value = tokeniser()
	try:
		return IP.create(value)
	except (IndexError,ValueError,socket.error):
		raise ValueError('"%s" is an invalid IP address' % value)



# encoding: utf-8
"""
keepalive.py

Created by Thomas Mangin on 2009-11-05.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from exabgp.util import character
from exabgp.util import hexstring

from exabgp.bgp.message.message import Message
from exabgp.bgp.message.notification import Notify

# =================================================================== KeepAlive
#


@Message.register
class KeepAlive (Message):
	ID = Message.CODE.KEEPALIVE
	TYPE = character(Message.CODE.KEEPALIVE)

	def message (self,negotiated=None):
		return self._message(b'')

	def __str__ (self):
		return "KEEPALIVE"

	@classmethod
	def unpack_message (cls, data, negotiated):  # pylint: disable=W0613
		# This can not happen at decode time as we check the length of the KEEPALIVE message
		# But could happen when calling the function programmatically
		if data:
			raise Notify('Keepalive can not have any payload but contains %s', hexstring(data))
		return cls()



# encoding: utf-8
"""
reactor/async.py

Created by Thomas Mangin on 2017-07-01.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from collections import deque

from exabgp.logger import Logger
from exabgp.vendoring import six

class ASYNC (object):
	LIMIT = 500

	def __init__ (self):
		self.logger = Logger()
		self._async = deque()

	def ready (self):
		return len(self._async) > 0

	def schedule (self, uid, command, callback):
		self.logger.debug('async | %s | %s' % (uid,command),'reactor')
		self._async.append((uid,callback))

	def clear (self, deluid=None):
		if not self._async:
			return
		if deluid is None:
			# We could delete all the generators just to be safe
			self._async = deque()
			return
		running = deque()
		for (uid,generator) in self._async:
			if uid != deluid:
				running.append((uid,generator))
		self._async = running

	def run (self):
		if not self.ready():
			return False

		length = range(min(len(self._async),self.LIMIT))
		uid, generator = self._async.popleft()

		for _ in length:
			try:
				six.next(generator)
				six.next(generator)
			except StopIteration:
				if not self._async:
					return False
				uid, generator = self._async.popleft()
			except KeyboardInterrupt:
				raise
			except Exception as exc:
				self.logger.error('async | %s | problem with function' % uid,'reactor')
				for line in str(exc).split('\n'):
					self.logger.error('async | %s | %s' % (uid,line),'reactor')

		self._async.appendleft((uid, generator))
		return True



# encoding: utf-8
"""
protection.py

Created by Evelio Vila on 2016-12-01.
Copyright (c) 2014-2017 Exa Networks. All rights reserved.
"""

from exabgp.bgp.message.notification import Notify
from exabgp.bgp.message.update.attribute.bgpls.linkstate import LINKSTATE, LsGenericFlags

#       0                   1
#       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5
#      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#      |Protection Cap |    Reserved   |
#      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#      https://tools.ietf.org/html/rfc5307 Sec 1.2
#      0x01  Extra Traffic
#      0x02  Unprotected
#      0x04  Shared
#      0x08  Dedicated 1:1
#      0x10  Dedicated 1+1
#      0x20  Enhanced
#      0x40  Reserved
#      0x80  Reserved

@LINKSTATE.register()
class LinkProtectionType(object):
	TLV = 1093

	def __init__ (self, protectionflags):
		self.protectionflags = protectionflags

	def __repr__ (self):
		return "Link protection mask: %s" % (self.protectionflags)

	@classmethod
	def unpack (cls,data,length):
		if length != 2:
			raise Notify(3,5, "Wrong size for protection type TLV")
		else:
			# We only care about the first octect
			protection_mask = LsGenericFlags.unpack(data[0:1],
				LsGenericFlags.LS_PROTECTION_MASK)
		return cls(protectionflags=protection_mask)

	def json (self,compact=None):
		return '"link-protection-flags": {}'.format(self.protectionflags.json())



#!/usr/bin/env python

from __future__ import print_function
import os
import sys
import time

# When the parent dies we are seeing continual newlines, so we only access so many before stopping
counter = 1

# sleep a little bit or we will never see the asm in the configuration file
# and the message received just before we go to the established loop will be printed twice
time.sleep(1)


while True:
	try:
		if counter % 2:
			print('announce route 10.0.%d.%d next-hop 1.2.3.4' % (counter >> 8, counter %256))
			sys.stdout.flush()
			time.sleep(1)
			print('withdraw route 10.0.%d.%d next-hop 1.2.3.4' % (counter >> 8, counter %256))
			sys.stdout.flush()
		else:
			print('announce route 2001:%d:: next-hop ::1' % counter) 
			sys.stdout.flush()
			time.sleep(1)
			print('withdraw route 2001:%d:: next-hop ::1' % counter) 
			sys.stdout.flush()

		counter += 1
	except KeyboardInterrupt:
		pass
	except IOError:
		break



# encoding: utf-8
"""
command.py

Created by Thomas Mangin on 2015-12-15.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""


class Command (object):
	callback = {
		'text': {},
		'json': {},
		'neighbor': {},
		'options': {}
	}

	functions = []

	@classmethod
	def register (cls, encoding, name, neighbor=True, options=None):
		if name not in cls.functions:
			cls.functions.append(name)
			cls.functions.sort(reverse=True)
			cls.callback['options'][name] = options

		def register (function):
			cls.callback['neighbor'][name] = neighbor
			cls.callback[encoding][name] = function
			function.func_name = name.replace(' ','_')
			return function

		return register



# encoding: utf-8
"""
debug.py

Created by Thomas Mangin on 2011-03-29.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

import os
import sys
from exabgp.util.panic import PANIC
from exabgp.util.panic import FOOTER


def bug_report (dtype, value, trace):
	sys.stdout.flush()
	sys.stderr.flush()

	print(PANIC)
	sys.stdout.flush()
	sys.stderr.flush()

	import traceback

	print("-- Traceback\n\n")
	traceback.print_exception(dtype,value,trace)

	from exabgp.logger import Logger
	logger = Logger()

	print("\n\n-- Configuration\n\n")
	print(logger.config())
	print("\n\n-- Logging History\n\n")
	print(logger.history())
	print("\n\n\n")

	print(FOOTER)
	sys.stdout.flush()
	sys.stderr.flush()


def intercept (dtype, value, trace):
	bug_report(dtype, value, trace)
	if os.environ.get('PDB',None) not in [None,'0','']:
		import pdb
		pdb.pm()


def setup_report ():
	sys.excepthook = intercept



# encoding: utf-8
"""
nexthop.py

Created by Thomas Mangin on 2009-11-05.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from exabgp.protocol.family import AFI
from exabgp.protocol.ip import IP
from exabgp.protocol.ip import NoNextHop
from exabgp.bgp.message.update.attribute.attribute import Attribute


# ================================================================== NextHop (3)

# The inheritance order is important and attribute MUST be first for the righ register to be called
# At least until we rename them to be more explicit

@Attribute.register()
class NextHop (Attribute,IP):
	ID = Attribute.CODE.NEXT_HOP
	FLAG = Attribute.Flag.TRANSITIVE
	CACHING = True
	SELF = False

	# XXX: This is a bad API, as it works on non-raw data
	def __init__ (self, string, packed=None):
		self.init(string,packed)

	def __eq__ (self, other):
		return \
			self.ID == other.ID and \
			self.FLAG == other.FLAG and \
			self._packed == other.ton()

	def __ne__ (self, other):
		return not self.__eq__(other)

	def ton (self, negotiated=None, afi=AFI.undefined):
		return self._packed

	def pack (self, negotiated=None):
		return self._attribute(self.ton())

	@classmethod
	def unpack (cls, data, negotiated=None):
		if not data:
			return NoNextHop
		return IP.unpack(data,NextHop)

	def __repr__ (self):
		return IP.__repr__(self)


class NextHopSelf (NextHop):
	SELF = True

	def __init__ (self, afi):
		self.afi = afi

	def __repr__ (self):
		return 'self'

	def ipv4 (self):
		return self.afi == AFI.ipv4

	def pack (self,negotiated):
		return self._attribute(negotiated.nexthopself(self.afi).ton())

	def ton (self, negotiated=None, afi=AFI.undefined):
		return negotiated.nexthopself(afi).ton()



#!/usr/bin/env python

import sys
import time
import random

def write (data):
	sys.stdout.write(data + '\n')
	sys.stdout.flush()

def main ():
	count = 0

	ip = {}
	nexthop="1.2.3.4"

	for ip1 in range(0,223):
		generated = '%d.0.0.0/8' % (ip1)
		ip[generated] = nexthop

	for ip1 in range(0,223):
		for ip2 in range(0,256):
			generated = '%d.%d.0.0/16' % (ip1,ip2)
			ip[generated] = nexthop

	# initial table dump
	for k,v in ip.iteritems():
		count += 1
		write('announce route %s next-hop %s med 100 as-path [ 100 101 102 103 104 105 106 107 108 109 110 ]' % (k,v))
		if count % 100 == 0:
			sys.stderr.write('initial : announced %d\n' % count)

	count &= 0xFFFFFFFe
	time.sleep(10000)

if __name__ == '__main__':
	try:
		main()
	except KeyboardInterrupt:
		pass



# encoding: utf-8
"""
rt.py

Created by Thomas Mangin on 2014-06-20.
Copyright (c) 2014-2017 Orange. All rights reserved.
Copyright (c) 2014-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from struct import pack
from struct import unpack

from exabgp.protocol.ip import IPv4
from exabgp.bgp.message.open.asn import ASN
from exabgp.bgp.message.update.attribute.community.extended import ExtendedCommunity


# ================================================================== RouteTarget
# RFC 4360 / RFC 7153

class RouteTarget (ExtendedCommunity):
	COMMUNITY_SUBTYPE = 0x02
	LIMIT = 0
	DESCRIPTION = 'target'

	@property
	def la (self):
		return self.community[2:self.LIMIT]

	@property
	def ga (self):
		return self.community[self.LIMIT:8]

	def __eq__ (self, other):
		return \
			self.COMMUNITY_SUBTYPE == other.COMMUNITY_SUBTYPE and \
			ExtendedCommunity.__eq__(self,other)

	def __ne__ (self, other):
		return not self.__eq__(other)


# ============================================================= RouteTargetASN2Number
# RFC 4360 / RFC 7153

@ExtendedCommunity.register
class RouteTargetASN2Number (RouteTarget):
	COMMUNITY_TYPE = 0x00
	LIMIT = 4

	__slots__ = ['asn','number']

	def __init__ (self, asn, number, transitive=True, community=None):
		self.asn = asn
		self.number = number
		# assert(number < pow(2,32))
		RouteTarget.__init__(
			self,
			community if community else pack(
				'!2sHL',
				self._subtype(transitive),
				asn,number
			)
		)

	def __hash__(self):
		return hash((self.asn,self.number))

	def __repr__ (self):
		return "%s:%d:%d" % (self.DESCRIPTION,self.asn,self.number)

	@classmethod
	def unpack (cls, data):
		asn,number = unpack('!HL',data[2:8])
		return cls(ASN(asn),number,False,data[:8])


# ============================================================= RouteTargetIPNumber
# RFC 4360 / RFC 7153

@ExtendedCommunity.register
class RouteTargetIPNumber (RouteTarget):
	COMMUNITY_TYPE = 0x01
	LIMIT = 6

	__slots__ = ['ip','number']

	def __init__ (self, ip, number, transitive=True, community=None):
		self.ip = ip
		self.number = number
		# assert(number < pow(2,16))
		RouteTarget.__init__(
			self,
			community if community else pack(
				'!2s4sH',
				self._subtype(transitive),
				IPv4.pton(ip),number
			)
		)

	# why could we not simply use ExtendedCommunity.hash ?
	def __hash__(self):
		return hash((self.ip,self.number))

	def __repr__ (self):
		return "%s:%s:%d" % (self.DESCRIPTION,self.ip,self.number)

	@classmethod
	def unpack (cls, data):
		ip,number = unpack('!4sH',data[2:8])
		return cls(IPv4.ntop(ip),number,False,data[:8])


# ======================================================== RouteTargetASN4Number
# RFC 4360 / RFC 7153

@ExtendedCommunity.register
class RouteTargetASN4Number (RouteTarget):
	COMMUNITY_TYPE = 0x02
	LIMIT = 6

	__slots__ = ['asn','number']

	def __init__ (self, asn, number, transitive=True, community=None):
		self.asn = asn
		self.number = number
		# assert(number < pow(2,16))
		RouteTarget.__init__(
			self,
			community if community else pack(
				'!2sLH',
				self._subtype(transitive),
				asn,number
			)
		)

	# why could we not simply use ExtendedCommunity.hash ?
	def __hash__(self):
		return hash((self.asn,self.number))

	def __repr__ (self):
		return "%s:%dL:%d" % (self.DESCRIPTION,self.asn, self.number)

	@classmethod
	def unpack (cls, data):
		asn,number = unpack('!LH',data[2:8])
		return cls(ASN(asn),number,False,data[:8])



"""
prefixv4.py

Created by Evelio Vila on 2016-11-26. eveliovila@gmail.com
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from struct import unpack

from exabgp.bgp.message.update.nlri.bgpls.nlri import BGPLS
from exabgp.bgp.message.update.nlri.bgpls.nlri import PROTO_CODES
from exabgp.bgp.message.update.nlri.bgpls.tlvs.node import NodeDescriptor
from exabgp.bgp.message.update.nlri.bgpls.tlvs.ospfroute import OspfRoute
from exabgp.bgp.message.update.nlri.bgpls.tlvs.ipreach import IpReach

#   The IPv4 and IPv6 Prefix NLRIs (NLRI Type = 3 and Type = 4) use the
#   same format, as shown in the following figure.
#
#      0                   1                   2                   3
#      0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
#     +-+-+-+-+-+-+-+-+
#     |  Protocol-ID  |
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     |                           Identifier                          |
#     |                            (64 bits)                          |
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     //              Local Node Descriptors (variable)              //
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     //                Prefix Descriptors (variable)                //
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+


@BGPLS.register
class PREFIXv4(BGPLS):
	CODE = 3
	NAME = " IPv4 Topology Prefix"
	SHORT_NAME = "PREFIX_V4"

	def __init__ (
			self,domain,proto_id,local_node,
			packed=None,ospf_type=None,prefix=None,
			nexthop=None,route_d=None,action=None, addpath=None):
		BGPLS.__init__(self,action,addpath)
		self.domain = domain
		self.ospf_type = ospf_type
		self.proto_id = proto_id
		self.local_node = local_node
		self.prefix = prefix
		self.nexthop = nexthop
		self._pack = packed
		self.route_d = route_d

	@classmethod
	def unpack_nlri (cls, data, rd):
		ospf_type = None
		proto_id = unpack('!B',data[0:1])[0]
		if proto_id not in PROTO_CODES.keys():
			raise Exception('Protocol-ID {} is not valid'.format(proto_id))
		domain = unpack('!Q',data[1:9])[0]
		tlvs = data[9:]

		while tlvs:
			tlv_type, tlv_length = unpack('!HH', tlvs[:4])
			if tlv_type == 256:
				values = tlvs[4: 4 + tlv_length]
				local_node = []
				while values:
					# Unpack Local Node Descriptor Sub-TLVs
					# We pass proto_id as TLV interpretation
					# follows IGP type
					node, left = NodeDescriptor.unpack(values, proto_id)
					local_node.append(node)
					if left == values:
						raise RuntimeError("sub-calls should consume data")
					values = left
				tlvs = tlvs[4 + tlv_length:]
				continue
			if tlv_type == 264:
				values = tlvs[4: 4 + tlv_length]
				ospf_type = OspfRoute.unpack(values)
				tlvs = tlvs[4 + tlv_length:]
			if tlv_type == 265:
				values = tlvs[4: 4 + tlv_length]
				prefix = IpReach.unpack(values)
				tlvs = tlvs[4 + tlv_length:]

		return cls(
			domain=domain,proto_id=proto_id,packed=data,
			local_node=local_node,ospf_type=ospf_type,
			prefix=prefix,route_d=rd
		)

	def __eq__ (self, other):
		return \
			isinstance(other, PREFIXv4) and \
			self.CODE == other.CODE and \
			self.domain == other.domain and \
			self.proto_id == other.proto_id and \
			self.route_d == other.route_d

	def __ne__ (self, other):
		return not self.__eq__(other)

	def __str__ (self):
		return self.json()

	def __hash__ (self):
		return hash((self))

	def json (self, compact=None):
		nodes = ', '.join(d.json() for d in self.local_node)
		content = ', '.join([
			'"ls-nlri-type": 3',
			'"l3-routing-topology": %d' % int(self.domain),
			'"protocol-id": %d' % int(self.proto_id),
			'"node-descriptors": { %s }' % nodes,
			self.prefix.json(),
			'"nexthop": "%s"' % self.nexthop,
		])
		if self.ospf_type:
			content += ', %s' % self.ospf_type.json()

		if self.route_d:
			content += ', %s' % self.route_d.json()

		return '{ %s }' % (content)

	def pack (self,negotiated=None):
		return self._pack



#!/usr/bin/env python
# encoding: utf-8
"""
cache.py

Created by David Farrar on 2012-12-27.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

import unittest

import time
from exabgp.util.cache import Cache
from exabgp.vendoring import six

from exabgp.vendoring.six.moves import xrange


class TestCache (unittest.TestCase):
	def test_speed (self):
		class klass1:
			def __init__ (self, data):
				pass

		class klass2 (object):
			def __init__ (self, data):
				pass

		class klass3:
			def __init__ (self, data):
				self.a = data[0]
				self.b = data[1]
				self.c = data[2]
				self.d = data[3]
				self.e = data[4]

		class klass4:
			def __init__ (self, data):
				self.a = data[0]
				self.b = data[1]
				self.c = data[2]
				self.d = data[3]
				self.e = data[4]

		class _kparent1:
			def __init__ (self, data):
				self.a = data[0]
				self.b = data[1]

		class _kparent2 (object):
			def __init__ (self, data):
				self.a = data[0]
				self.b = data[1]

		class klass5 (_kparent1):
			def __init__ (self, data):
				_kparent1.__init__(self,data)
				self.c = data[2]
				self.d = data[3]
				self.e = data[4]

		class klass6 (_kparent2):
			def __init__ (self, data):
				_kparent2.__init__(self,data)
				self.c = data[2]
				self.d = data[3]
				self.e = data[4]

		class klass7 (klass6):
			pass

		class klass8 (klass6):
			def __init__ (self, data):
				klass6.__init__(self,data)
				self.s = self.a + self.b + self.c + self.d + self.e

		class klass9 (klass6):
			def __init__ (self, data):
				klass6.__init__(self,data)
				self.s1 = self.a + self.b + self.c + self.d + self.e
				self.s2 = self.b + self.c + self.d + self.e
				self.s3 = self.c + self.d + self.e
				self.s4 = self.d + self.e
				self.s5 = self.a + self.b + self.c + self.d
				self.s6 = self.a + self.b + self.c
				self.s7 = self.a + self.b

		COUNT = 100000
		UNIQUE = 5000

		samples = set()
		chars = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789:"|;<>?,./[]{}-=_+!@£$%^&*()'

		from random import choice

		while len(samples) != UNIQUE:
			samples.add(choice(chars)+choice(chars)+choice(chars)+choice(chars)+choice(chars))

		samples = list(samples)

		for klass in [klass1,klass2,klass3,klass4,klass5,klass6,klass7,klass8,klass9]:
			cache = {}

			start = time.time()
			for val in xrange(COUNT):
				val %= UNIQUE
				_ = klass(samples[val])
			end = time.time()
			time1 = end-start

			print(COUNT,'iterations of',klass.__name__,'with',UNIQUE,'uniques classes')
			print("time instance %d" % time1)

			cache = Cache()
			start = time.time()
			for val in xrange(COUNT):
				val %= UNIQUE

				if val in cache:
					_ = cache.retrieve(val)
				else:
					_ = cache.cache(val, klass(samples[val]))

			end = time.time()
			time2 = end-start

			print("time cached  %d" % time2)
			print("speedup %.3f" % (time1/time2))
			print()



# encoding: utf-8
"""
neighaddr.py

Created by Evelio Vila on 2016-11-26. eveliovila@gmail.com
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from struct import pack
from struct import unpack

from exabgp.protocol.ip import IP
from exabgp.bgp.message.notification import Notify
from exabgp.util import ordinal

#  https://tools.ietf.org/html/rfc5305#section-3.3
#   This sub-TLV contains a single IPv4 address for a neighboring router
#   on this link.  This sub-TLV can occur multiple times.
#
#   Implementations MUST NOT inject a /32 prefix for the neighbor address
#   into their routing or forwarding table because this can lead to
#   forwarding loops when interacting with systems that do not support
#   this sub-TLV.
# ================================================================== NeighborAddress


class NeighAddr (object):
	def __init__ (self, addr, packed=None):
		self.addr = addr
		self._packed = packed

	@classmethod
	def unpack (cls, data):
		if len(data) == 4:
			# IPv4 address
			addr = IP.unpack(data[:4])
		elif len(data) == 16:
			# IPv6
			addr = IP.unpack(data[:16])
		return cls(addr=addr)

	def json (self):
		content = ' '.join([
				'"neighbor-address": "%s"' % self.addr,
		])
		return content

	def __eq__ (self, other):
		return self.addr == other.addr

	def __neq__ (self, other):
		return self.addr != other.addr

	def __lt__ (self, other):
		raise RuntimeError('Not implemented')

	def __le__ (self, other):
		raise RuntimeError('Not implemented')

	def __gt__ (self, other):
		raise RuntimeError('Not implemented')

	def __ge__ (self, other):
		raise RuntimeError('Not implemented')

	def __str__ (self):
		return ':'.join('%02X' % ordinal(_) for _ in self._packed)

	def __repr__ (self):
		return self.__str__()

	def __len__ (self):
		return len(self._packed)

	def __hash__ (self):
		return hash(str(self))

	def pack (self):
		return self._packed



# encoding: utf-8
"""
sralgo.py

Created by Evelio Vila
Copyright (c) 2014-2017 Exa Networks. All rights reserved.
"""

import json
from struct import unpack
from exabgp.vendoring import six

from exabgp.bgp.message.update.attribute.bgpls.linkstate import LINKSTATE

#     draft-gredler-idr-bgp-ls-segment-routing-ext-03
#    0                   1                   2                   3
#    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
#   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#   |            Type               |            Length             |
#   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#   |  Algorithm 1  |  Algorithm... |  Algorithm N |                |
#   +-                                                             -+
#   |                                                               |
#   +                                                               +
#						sec 2.1.2.

@LINKSTATE.register()
class SrAlgorithm(object):
	TLV = 1035

	def __init__ (self, sr_algos):
		self.sr_algos = sr_algos

	def __repr__ (self):
		return "SrAlgorithms: %s" % (self.sr_algos)

	@classmethod
	def unpack (cls,data,length):
		sr_algos = []
		while data:
			sr_algos.append(six.indexbytes(data,0))
			data = data[1:]
		# Looks like IOS XR advertises len 0 on this sub TLV
		# when using default SPF.
		if not len(sr_algos):
			sr_algos.append(0)
		return cls(sr_algos=sr_algos)

	def json (self,compact=None):
		return '"sr-algorithms": {}'.format(json.dumps(self.sr_algos))



"""
ethernetad.py

Created by Thomas Mangin on 2014-06-27.
Copyright (c) 2014-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

# from struct import pack
# from struct import unpack

from exabgp.util import concat_bytes

# from exabgp.protocol.family import AFI
# from exabgp.protocol.family import SAFI
# from exabgp.bgp.message.update.nlri.qualifier import ESI

from exabgp.bgp.message.update.nlri.qualifier import RouteDistinguisher
from exabgp.bgp.message.update.nlri.qualifier import Labels
from exabgp.bgp.message.update.nlri.qualifier import ESI
from exabgp.bgp.message.update.nlri.qualifier import EthernetTag

from exabgp.bgp.message.update.nlri.evpn.nlri import EVPN

# +---------------------------------------+
# |      RD   (8 octets)                  |
# +---------------------------------------+
# |Ethernet Segment Identifier (10 octets)|
# +---------------------------------------+
# |  Ethernet Tag ID (4 octets)           |
# +---------------------------------------+
# |  MPLS Label (3 octets)                |
# +---------------------------------------+

# ===================================================================== EVPNNLRI


@EVPN.register
class EthernetAD (EVPN):
	CODE = 1
	NAME = "Ethernet Auto-Discovery"
	SHORT_NAME = "EthernetAD"

	def __init__ (self, rd, esi, etag, label, packed=None,nexthop=None,action=None,addpath=None):
		EVPN.__init__(self,action,addpath)
		self.nexthop = nexthop
		self.rd = rd
		self.esi = esi
		self.etag = etag
		self.label = label if label else Labels.NOLABEL
		self._pack(packed)

	def __eq__ (self, other):
		return \
			isinstance(other, EthernetAD) and \
			self.CODE == other.CODE and \
			self.rd == other.rd and \
			self.etag == other.etag
		# esi and label must not be part of the comparaison

	def __ne__ (self, other):
		return not self.__eq__(other)

	def __str__ (self):
		return "%s:%s:%s:%s:%s" % (
			self._prefix(),
			self.rd._str(),
			self.esi,
			self.etag,
			self.label
		)

	def __hash__ (self):
		# esi and label MUST *NOT* be part of the hash
		return hash((self.rd,self.etag))

	def _pack (self, packed=None):
		if self._packed:
			return self._packed

		if packed:
			self._packed = packed
			return packed

		self._packed = concat_bytes(
			self.rd.pack(),
			self.esi.pack(),
			self.etag.pack(),
			self.label.pack()
		)
		return self._packed

	@classmethod
	def unpack (cls, data):
		rd = RouteDistinguisher.unpack(data[:8])
		esi = ESI.unpack(data[8:18])
		etag = EthernetTag.unpack(data[18:22])
		label = Labels.unpack(data[22:25])

		return cls(rd,esi,etag,label,data)

	def json (self, compact=None):
		content = ' "code": %d, ' % self.CODE
		content += '"parsed": true, '
		content += '"raw": "%s", ' % self._raw()
		content += '"name": "%s", ' % self.NAME
		content += '%s, ' % self.rd.json()
		content += '%s, ' % self.esi.json()
		content += '%s, ' % self.etag.json()
		content += '%s ' % self.label.json()
		return '{%s}' % content



#!/usr/bin/env python
# encoding: utf-8
"""
data.py

Created by Thomas Mangin on 2009-09-06.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

import unittest

from exabgp.configuration.environment import environment
environment.setup('')

# from exabgp.protocol.family import AFI
# from exabgp.protocol.family import SAFI

class TestData (unittest.TestCase):
	def setUp (self):
		pass

	# def test_1_nlri_1 (self):
	# 	self.assertEqual(''.join([chr(c) for c in [32,1,2,3,4]]),to_NLRI('1.2.3.4','32').pack())
	# def test_1_nlri_2 (self):
	# 	self.assertEqual(''.join([chr(c) for c in [24,1,2,3]]),to_NLRI('1.2.3.4','24').pack())
	# def test_1_nlri_3 (self):
	# 	self.assertEqual(''.join([chr(c) for c in [20,1,2,3]]),to_NLRI('1.2.3.4','20').pack())
	#
	# def test_2_ip_2 (self):
	# 	self.assertEqual(str(InetIP('::ffff:192.168.1.26')),'::ffff:192.168.1.26/128')
	# 	self.assertEqual(str(InetIP('::ffff:192.168.1.26').ip),'::ffff:192.168.1.26')
	#
	# def test_3_ipv6_1 (self):
	# 	default = InetIP('::')
	# 	self.assertEqual(str(default),'::/128')
	# 	self.assertEqual(default.ip,'::')
	# 	self.assertEqual(default.packedip(),'\0'*16)
	#
	# def test_3_ipv6_2 (self):
	# 	default = InetIP('1234:5678::')
	# 	self.assertEqual(str(default),'1234:5678::/128')
	# 	self.assertEqual(default.ip,'1234:5678::')
	# 	self.assertEqual(default.packedip(),'\x12\x34\x56\x78'+'\0'*12)
	#
	# def test_3_ipv6_3 (self):
	# 	default = InetIP('1234:5678::1')
	# 	self.assertEqual(str(default),'1234:5678::1/128')
	# 	self.assertEqual(default.ip,'1234:5678::1')
	# 	self.assertEqual(default.packedip(),'\x12\x34\x56\x78'+'\0'*11 + '\x01')
	#
	# def test_xxx (self):
	# 	ip = "192.0.2.0"
	# 	net  = chr (192) + chr(0) + chr(2) +chr(0)
	# 	bnt = chr(24) + chr (192) + chr(0) + chr(2)
	#
	# 	pfx = Prefix(AFI.ipv4,ip,24)
	# 	bgp = BGPNLRI(AFI.ipv4,bnt)
	#
	# 	self.assertEqual(str(pfx),"%s/24" % ip)
	# 	self.assertEqual(str(afi),"%s/24" % ip)
	# 	self.assertEqual(str(bgp),"%s/24" % ip)
	#
	# 	self.assertEqual(pfx.pack(),bnt)
	# 	self.assertEqual(afi.pack(),bnt)
	# 	self.assertEqual(bgp.pack(),bnt)
	#
	# 	# README: NEED To add ASN test
	# 	# README: NEED To add NLRI test

if __name__ == '__main__':
	unittest.main()



# encoding: utf-8
"""
__init__.py

Created by Thomas Mangin on 2015-05-15.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

import string
import sys

PY2 = sys.version_info[0] < 3


def hexstring (value):
	def spaced (value):
		for v in value:
			yield '%02X' % ordinal(v)
	return '0x' + concat_strs_i(spaced(value))


def hexbytes (value):
	return bytes_ascii(hexstring(str_ascii(value)))


def string_is_hex (s):
	if s[:2].lower() != '0x':
		return False
	if len(s) <= 2:
		return False
	return all(c in string.hexdigits for c in s[2:])

# for Python3+, let's redefine ord into something
# that plays along nicely with ord(data[42]) with
# data being of type 'bytes'


if PY2:
	ordinal = ord
else:
	def ordinal(x):
		return x if type(x) == int else ord(x)


if PY2:
	character = chr
else:
	def character(x):
		return bytes([x])


if PY2:
	def padding(n):
		return '\0'*n
else:
	def padding(n):
		return bytes(n)


# Each item is an 'str' in py2 or a 'bytes' in py3

def concat_strs(*items):
	return ''.join(items)

def concat_bytes(*items):
	return b''.join(items)

# same with iterators/lists

def concat_strs_i(iterable):
	return ''.join(iterable)

def concat_bytes_i(iterable):
	return b''.join(iterable)

# helpers for converting between string and bytestring

if PY2:
	def str_ascii (string):
		return string
else:
	def str_ascii(bytestring):
		return str(bytestring, 'ascii')

if PY2:
	def bytes_ascii (string):
		return string
else:
	def bytes_ascii(bytestring):
		return bytes(bytestring, 'ascii')



# encoding: utf-8
"""
igpmetric.py

Created by Evelio Vila on 2016-12-01.
Copyright (c) 2014-2017 Exa Networks. All rights reserved.
"""

from struct import unpack
from exabgp.vendoring import six

from exabgp.vendoring.bitstring import BitArray
from exabgp.bgp.message.notification import Notify

from exabgp.bgp.message.update.attribute.bgpls.linkstate import LINKSTATE


#   The IGP Metric TLV carries the metric for this link.  The length of
#   this TLV is variable, depending on the metric width of the underlying
#   protocol.  IS-IS small metrics have a length of 1 octet (the two most
#   significant bits are ignored).  OSPF link metrics have a length of 2
#   octets.  IS-IS wide metrics have a length of 3 octets.
#
#      0                   1                   2                   3
#      0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     |              Type             |             Length            |
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     //      IGP Link Metric (variable length)      //
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

@LINKSTATE.register()
class IgpMetric(object):
	TLV = 1095

	def __init__ (self, igpmetric):
		self.igpmetric = igpmetric

	def __repr__ (self):
		return "IGP Metric: %s" % (self.igpmetric)

	@classmethod
	def unpack (cls,data,length):
		if len(data) == 2:
			# OSPF
			igpmetric = unpack('!H',data)[0]
			return cls(igpmetric=igpmetric)
		elif len(data) == 1:
			# ISIS small metrics
			igpmetric = six.indexbytes(data,0)
			return cls(igpmetric=igpmetric)
		elif len(data) == 3:
			# ISIS wide metrics
			b = BitArray(bytes=data)
			igpmetric = b.unpack('uintbe:24')
			return cls(igpmetric=igpmetric)
		else:
			raise Notify(3,5, "Incorrect IGP Metric Size")

	def json (self,compact=None):
		return '"igp-metric": %d' % int(self.igpmetric[0])



#!/usr/bin/env python
# encoding: utf-8
"""
l2vpn.py

Created by Thomas Mangin on 2010-01-14.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

import unittest

from exabgp.bgp.message.update.nlri import VPLS
from exabgp.bgp.message.update.nlri.qualifier import RouteDistinguisher


class TestL2VPN (unittest.TestCase):
	@staticmethod
	def generate_rd (rd):
		"""only ip:num is supported atm.code from configure.file"""
		separator = rd.find(':')
		prefix = rd[:separator]
		suffix = int(rd[separator+1:])
		data = [chr(0),chr(1)]
		data.extend([chr(int(_)) for _ in prefix.split('.')])
		data.extend([chr(suffix >> 8),chr(suffix & 0xFF)])
		bin_rd = ''.join(data)
		return RouteDistinguisher(bin_rd)

	def setUp (self):
		"""
		setUp unittesting

		l2vpn:endpoint:3:base:262145:offset:1:size:8: route-distinguisher 172.30.5.4:13
		l2vpn:endpoint:3:base:262145:offset:1:size:8: route-distinguisher 172.30.5.3:11
		"""
		self.encoded_l2vpn_nlri1 = bytearray.fromhex(u'0011 0001 AC1E 0504 000D 0003 0001 0008 4000 11')
		self.encoded_l2vpn_nlri2 = bytearray.fromhex(u'0011 0001 AC1E 0503 000B 0003 0001 0008 4000 11')
		self.decoded_l2vpn_nlri1 = VPLS(TestL2VPN.generate_rd('172.30.5.4:13'),3,262145,1,8)
		self.decoded_l2vpn_nlri2 = VPLS(TestL2VPN.generate_rd('172.30.5.3:11'),3,262145,1,8)
		"""
		output from Juniper
		Communities: target:54591:6 Layer2-info: encaps: VPLS, control flags:[0x0] , mtu: 0, site preference: 100
		"""
		self.encoded_ext_community = bytearray.fromhex(u'0002 D53F 0000 0006 800A 1300 0000 0064')

	def test_l2vpn_decode (self):
		"""
		decode and test against known data

		we do know what routes Juniper sends us and we testing decoded values against it
		"""
		# l2vpn_route1 = VPLS.unpack(str(self.encoded_l2vpn_nlri1))
		# l2vpn_route2 = VPLS.unpack(str(self.encoded_l2vpn_nlri2))
		# self.assertEqual(l2vpn_route1.endpoint,3)
		# self.assertEqual(l2vpn_route1.rd._str(),'172.30.5.4:13')
		# self.assertEqual(l2vpn_route1.offset,1)
		# self.assertEqual(l2vpn_route1.base,262145)
		# self.assertEqual(l2vpn_route1.size,8)
		# self.assertEqual(l2vpn_route2.endpoint,3)
		# self.assertEqual(l2vpn_route2.rd._str(),'172.30.5.3:11')
		# self.assertEqual(l2vpn_route2.offset,1)
		# self.assertEqual(l2vpn_route2.base,262145)
		# self.assertEqual(l2vpn_route2.size,8)

	def test_l2vpn_encode (self):
		"""
		encode and test against known data

		we are encoding routes and testing em against what we have recvd from Juniper
		"""
		# encoded_l2vpn = VPLS(None,None,None,None,None)
		# encoded_l2vpn = self.decoded_l2vpn_nlri1
		# self.assertEqual(
		# 	encoded_l2vpn.pack().encode('hex'),
		# 	str(self.encoded_l2vpn_nlri1).encode('hex')
		# )
		#
		# encoded_l2vpn.nlri = self.decoded_l2vpn_nlri2
		# encoded_l2vpn.rd = self.decoded_l2vpn_nlri2.rd
		# self.assertEqual(
		# 	encoded_l2vpn.pack().encode('hex'),
		# 	str(self.encoded_l2vpn_nlri2).encode('hex')
		# )

	# Disable until we refactor the configuation code
	#
	# def test_l2info_community_decode (self):
	# 	'''
	# 	Juniper sends us both target and l2info; so we test only against
	# 	l2info.
	# 	'''
	# 	l2info_com = ExtendedCommunity.unpack(str(self.encoded_ext_community)[8:16])
	# 	self.assertEqual(l2info_com,to_ExtendedCommunity('l2info:19:0:0:100'))

	# def test_l2info_community_encode (self):
	# 	l2info_com_encoded = to_ExtendedCommunity('l2info:19:0:0:100')
	# 	self.assertEqual(l2info_com_encoded.pack(),str(self.encoded_ext_community)[8:16])

if __name__ == '__main__':
		unittest.main()



# encoding: utf-8
"""
hostname.py

Created by Thomas Mangin on 2015-05-16.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

# https://tools.ietf.org/html/draft-walton-bgp-hostname-capability-02


from exabgp.util import character
from exabgp.util import ordinal
from exabgp.util import concat_bytes
from exabgp.bgp.message.open.capability.capability import Capability


class HostName (Capability):
	ID = Capability.CODE.HOSTNAME

	def __init__ (self, host_name, domain_name):
		self.host_name = host_name
		self.domain_name = domain_name

	def __str__ (self):
		return 'Hostname(%s %s)' % (self.host_name,self.domain_name)

	def json (self):
		return '{ "host-name": "%s", "domain-name": "%s" }' % (self.host_name,self.domain_name)

	def extract (self):
		return [concat_bytes(
			character(len(self.host_name)),
			self.host_name,
			character(len(self.domain_name)),
			self.domain_name,
		)]

	@staticmethod
	def unpack_capability (instance, data, capability=None):  # pylint: disable=W0613
		l1 = ordinal(data[0])
		instance.host_name = data[1:l1+1].decode('utf-8')
		l2 = ordinal(data[l1+1])
		instance.domain_name = data[l1+2:l1+2+l2].decode('utf-8')
		return instance



# encoding: utf-8
"""
attributes.py

Created by Thomas Mangin on 2009-11-05.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from struct import unpack

from exabgp.configuration.environment import environment

from exabgp.util import ordinal
from exabgp.bgp.message.update.attribute.attribute import Attribute
from exabgp.bgp.message.update.attribute.attribute import TreatAsWithdraw
from exabgp.bgp.message.update.attribute.attribute import Discard
from exabgp.bgp.message.update.attribute.generic import GenericAttribute
from exabgp.bgp.message.update.attribute.origin import Origin
from exabgp.bgp.message.update.attribute.aspath import ASPath
from exabgp.bgp.message.update.attribute.localpref import LocalPreference

# For bagpipe
from exabgp.bgp.message.update.attribute.community import Communities

from exabgp.bgp.message.notification import Notify

from exabgp.util import ordinal

from exabgp.logger import Logger
from exabgp.logger import LazyAttribute

from exabgp.vendoring import six


class _NOTHING (object):
	def pack (self, _=None):
		return b''


NOTHING = _NOTHING()


# =================================================================== Attributes
#

# 0                   1
# 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |  Attr. Flags  |Attr. Type Code|
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

class Attributes (dict):
	INTERNAL = (
		Attribute.CODE.INTERNAL_SPLIT,
		Attribute.CODE.INTERNAL_WATCHDOG,
		Attribute.CODE.INTERNAL_NAME,
		Attribute.CODE.INTERNAL_WITHDRAW,
		# Attribute.CODE.INTERNAL_DISCARD,
		# Attribute.CODE.INTERNAL_TREAT_AS_WITHDRAW,
	)

	NO_GENERATION = (
		Attribute.CODE.NEXT_HOP,
	) + INTERNAL

	TREAT_AS_WITHDRAW = (
		Attribute.CODE.ORIGIN,
		Attribute.CODE.AS_PATH,
		Attribute.CODE.NEXT_HOP,
		Attribute.CODE.MED,
		Attribute.CODE.LOCAL_PREF,
		Attribute.CODE.LARGE_COMMUNITY
	)

	DISCARD = (
		Attribute.CODE.ATOMIC_AGGREGATE,
		Attribute.CODE.AGGREGATOR,
	)

	MANDATORY = (
		Attribute.CODE.ORIGIN,
		Attribute.CODE.AS_PATH,
		Attribute.CODE.LOCAL_PREF
	)

	NO_DUPLICATE = (
		Attribute.CODE.MP_REACH_NLRI,
		Attribute.CODE.MP_UNREACH_NLRI,
	)

	VALID_ZERO = (
		Attribute.CODE.ATOMIC_AGGREGATE,
		Attribute.CODE.AS_PATH,
	)

	# A cache of parsed attributes
	cache = {}

	# The previously parsed Attributes
	cached = None
	# previously parsed attribute, from which cached was made of
	previous = ''

	representation = {
		# key:  (how, default, name, text_presentation, json_presentation),
		Attribute.CODE.ORIGIN:             ('string',  '', 'origin',             '%s',     '%s'),
		Attribute.CODE.AS_PATH:            ('multiple','', ('as-path','as-set','confederation-path','confederation-set'), '%s',     '%s'),
		Attribute.CODE.NEXT_HOP:           ('string',  '', 'next-hop',           '%s',     '%s'),
		Attribute.CODE.MED:                ('integer', '', 'med',                '%s',     '%s'),
		Attribute.CODE.LOCAL_PREF:         ('integer', '', 'local-preference',   '%s',     '%s'),
		Attribute.CODE.ATOMIC_AGGREGATE:   ('boolean', '', 'atomic-aggregate',   '%s',     '%s'),
		Attribute.CODE.AGGREGATOR:         ('string',  '', 'aggregator',         '( %s )', '%s'),
		Attribute.CODE.AS4_AGGREGATOR:     ('string',  '', 'aggregator',         '( %s )', '%s'),
		Attribute.CODE.COMMUNITY:          ('list',    '', 'community',          '%s',     '%s'),
		Attribute.CODE.LARGE_COMMUNITY:    ('list',    '', 'large-community',    '%s',     '%s'),
		Attribute.CODE.ORIGINATOR_ID:      ('inet',    '', 'originator-id',      '%s',     '%s'),
		Attribute.CODE.CLUSTER_LIST:       ('list',    '', 'cluster-list',       '%s',     '%s'),
		Attribute.CODE.EXTENDED_COMMUNITY: ('list',    '', 'extended-community', '%s',     '%s'),
		Attribute.CODE.PMSI_TUNNEL:        ('string',  '', 'pmsi',               '%s',     '%s'),
		Attribute.CODE.AIGP:               ('integer', '', 'aigp',               '%s',     '%s'),
		Attribute.CODE.BGP_LS:             ('list',    '', 'bgp-ls',             '%s',     '%s'),
		Attribute.CODE.BGP_PREFIX_SID:	   ('list',    '', 'bgp-prefix-sid',     '%s',     '%s'),
		Attribute.CODE.INTERNAL_NAME:      ('string',  '', 'name',               '%s',     '%s'),
		Attribute.CODE.INTERNAL_DISCARD:   ('string',  '', 'error',              '%s',     '%s'),
		Attribute.CODE.INTERNAL_TREAT_AS_WITHDRAW: ('string','','error',         '%s',     '%s'),
	}

	def _generate_text (self):
		for code in sorted(self.keys()):
			# XXX: FIXME: really we should have a INTERNAL attribute in the classes
			if code in Attributes.NO_GENERATION:
				continue

			attribute = self[code]

			if code not in self.representation:
				yield ' attribute [ 0x%02X 0x%02X %s ]' % (code,attribute.FLAG,str(attribute))
				continue

			if attribute.GENERIC:
				yield ' attribute [ 0x%02X 0x%02X %s ]' % (code,attribute.FLAG,str(attribute))
				continue

			how, _, name, presentation, _ = self.representation[code]
			if how == 'boolean':
				yield ' %s' % name
			elif how == 'list':
				yield ' %s %s' % (name, presentation % str(attribute))
			elif how == 'multiple':
				yield ' %s %s' % (name[0], presentation % str(attribute))
			else:
				yield ' %s %s' % (name, presentation % str(attribute))

	def _generate_json (self):
		for code in sorted(self.keys()):
			# remove the next-hop from the attribute as it is define with the NLRI
			if code in Attributes.NO_GENERATION:
				continue

			attribute = self[code]

			if code not in self.representation:
				yield '"attribute-0x%02X-0x%02X": "%s"' % (code,attribute.FLAG,str(attribute))
				continue

			how, _, name, _, presentation = self.representation[code]
			if how == 'boolean':
				yield '"%s": %s' % (name, 'true' if self.has(code) else 'false')
			elif how == 'string':
				yield '"%s": "%s"' % (name, presentation % str(attribute))
			elif how == 'list':
				yield '"%s": %s' % (name, presentation % attribute.json())
			elif how == 'multiple':
				for n in name:
					value = attribute.json(n)
					if value:
						yield '"%s": %s' % (n, presentation % value)
			elif how == 'inet':
				yield '"%s": "%s"' % (name, presentation % str(attribute))
			# Should never be ran
			else:
				yield '"%s": %s' % (name, presentation % str(attribute))

	def __init__ (self):
		dict.__init__(self)
		# cached representation of the object
		self._str = ''
		self._idx = ''
		self._json = ''
		# The parsed attributes have no mp routes and/or those are last
		self.cacheable = True

		# XXX: FIXME: surely not the best place for this
		Attribute.caching = environment.settings().cache.attributes

	def has (self, k):
		return k in self

	def add (self, attribute, _=None):
		# we return None as attribute if the unpack code must not generate them
		if attribute is None:
			return
		if attribute.ID in self:
			return

		self._str = ''
		self._json = ''

		self[attribute.ID] = attribute

	# This is as when we generate flow spec we can have multiple keywords
	# which are all adding information in the extended-community
	def add_and_merge (self, attribute):
		if attribute.ID not in self:
			self.add(attribute)
			return

		if attribute.ID == Attribute.CODE.EXTENDED_COMMUNITY:
			for community in attribute.communities:
				self[attribute.ID].add(community)

	def remove (self, attrid):
		self.pop(attrid)

	def watchdog (self):
		return self.pop(Attribute.CODE.INTERNAL_WATCHDOG,None)

	def withdraw (self):
		return self.pop(Attribute.CODE.INTERNAL_WITHDRAW,None) is not None

	def pack (self, negotiated, with_default=True):
		local_asn = negotiated.local_as
		peer_asn = negotiated.peer_as

		message = b''

		default = {
			Attribute.CODE.ORIGIN: lambda left,right: Origin(Origin.IGP),
			Attribute.CODE.AS_PATH: lambda left,right: ASPath([],[]) if left == right else ASPath([local_asn,],[]),
			Attribute.CODE.LOCAL_PREF: lambda left,right: LocalPreference(100) if left == right else NOTHING,
		}

		skip = {
			Attribute.CODE.NEXT_HOP: lambda left,right,nh: nh.ipv4() is not True,
			Attribute.CODE.LOCAL_PREF: lambda left,right,nh: left != right,
		}

		keys = list(self)
		alls = set(keys + list(default) if with_default else [])

		for code in sorted(alls):
			if code in Attributes.INTERNAL:
				continue

			if code not in keys and code in default:
				message += default[code](local_asn,peer_asn).pack(negotiated)
				continue

			attribute = self[code]

			if code in skip and skip[code](local_asn,peer_asn,attribute):
				continue

			message += attribute.pack(negotiated)

		return message

	def json (self):
		if not self._json:
			self._json = ', '.join(self._generate_json())
		return self._json

	def __repr__ (self):
		if not self._str:
			self._str = ''.join(self._generate_text())
		return self._str

	def index (self):
		# XXX: something a little bit smaller memory wise ?
		if not self._idx:
			idx = ''.join(self._generate_text())
			nexthop = str(self.get(Attribute.CODE.NEXT_HOP,b''))
			self._idx = '%s next-hop %s' % (idx,nexthop) if nexthop else idx
		return self._idx

	@classmethod
	def unpack (cls, data, negotiated):
		if cls.cached and data == cls.previous:
			return cls.cached

		attributes = cls().parse(data,negotiated)

		if Attribute.CODE.INTERNAL_TREAT_AS_WITHDRAW in attributes:
			return attributes

		if Attribute.CODE.AS_PATH in attributes and Attribute.CODE.AS4_PATH in attributes:
			attributes.merge_attributes()

		if Attribute.CODE.MP_REACH_NLRI not in attributes and Attribute.CODE.MP_UNREACH_NLRI not in attributes:
			cls.previous = data
			cls.cached = attributes
		else:
			cls.previous = ''
			cls.cached = None

		return attributes

	@staticmethod
	def flag_attribute_content (data):
		flag = Attribute.Flag(ordinal(data[0]))
		attr = Attribute.CODE(ordinal(data[1]))

		if flag & Attribute.Flag.EXTENDED_LENGTH:
			length = unpack('!H',data[2:4])[0]
			return flag, attr, data[4:length+4]
		else:
			length = ordinal(data[2])
			return flag, attr, data[3:length+3]

	def parse (self, data, negotiated):
		if not data:
			return self

		try:
			# We do not care if the attribute are transitive or not as we do not redistribute
			flag = Attribute.Flag(ordinal(data[0]))
			aid = Attribute.CODE(ordinal(data[1]))
		except IndexError:
			self.add(TreatAsWithdraw())
			return self

		try:
			offset = 3
			length = ordinal(data[2])

			if flag & Attribute.Flag.EXTENDED_LENGTH:
				offset = 4
				length = (length << 8) + ordinal(data[3])
		except IndexError:
			self.add(TreatAsWithdraw(aid))
			return self

		data = data[offset:]
		left = data[length:]
		attribute = data[:length]

		logger = Logger()
		logger.debug(LazyAttribute(flag,aid,length,data[:length]),'parser')

		# remove the PARTIAL bit before comparaison if the attribute is optional
		if aid in Attribute.attributes_optional:
			flag &= Attribute.Flag.MASK_PARTIAL & 0xFF
			# flag &= ~Attribute.Flag.PARTIAL & 0xFF  # cleaner than above (python use signed integer for ~)

		if aid in self:
			if aid in self.NO_DUPLICATE:
				raise Notify(3,1,'multiple attribute for %s' % str(Attribute.CODE(attribute.ID)))

			logger.debug('duplicate attribute %s (flag 0x%02X, aid 0x%02X) skipping' % (Attribute.CODE.names.get(aid,'unset'),flag,aid),'parser')
			return self.parse(left,negotiated)

		# handle the attribute if we know it
		if Attribute.registered(aid,flag):
			if length == 0 and aid not in self.VALID_ZERO:
				self.add(TreatAsWithdraw(aid))
				return self.parse(left,negotiated)

			try:
				decoded = Attribute.unpack(aid,flag,attribute,negotiated)
			except IndexError as exc:
				if aid in self.TREAT_AS_WITHDRAW:
					decoded = TreatAsWithdraw(aid)
				else:
					raise exc
			except Notify as exc:
				if aid in self.TREAT_AS_WITHDRAW:
					decoded = TreatAsWithdraw()
				elif aid in self.DISCARD:
					decoded = Discard()
				else:
					raise exc
			self.add(decoded)
			return self.parse(left,negotiated)

		# XXX: FIXME: we could use a fallback function here like capability

		# if we know the attribute but the flag is not what the RFC says.
		if aid in Attribute.attributes_known:
			if aid in self.TREAT_AS_WITHDRAW:
				logger.debug('invalid flag for attribute %s (flag 0x%02X, aid 0x%02X) treat as withdraw' % (Attribute.CODE.names.get(aid,'unset'),flag,aid),'parser')
				self.add(TreatAsWithdraw())
			if aid in self.DISCARD:
				logger.debug('invalid flag for attribute %s (flag 0x%02X, aid 0x%02X) discard' % (Attribute.CODE.names.get(aid,'unset'),flag,aid),'parser')
				return self.parse(left,negotiated)
			# XXX: Check if we are missing any
			logger.debug('invalid flag for attribute %s (flag 0x%02X, aid 0x%02X) unspecified (should not happen)' % (Attribute.CODE.names.get(aid,'unset'),flag,aid),'parser')
			return self.parse(left,negotiated)

		# it is an unknown transitive attribute we need to pass on
		if flag & Attribute.Flag.TRANSITIVE:
			logger.debug('unknown transitive attribute (flag 0x%02X, aid 0x%02X)' % (flag,aid),'parser')
			try:
				decoded = GenericAttribute(aid,flag | Attribute.Flag.PARTIAL,attribute)
			except IndexError:
				decoded = TreatAsWithdraw(aid)
			self.add(decoded,attribute)
			return self.parse(left,negotiated)

		# it is an unknown non-transitive attribute we can ignore.
		logger.debug('ignoring unknown non-transitive attribute (flag 0x%02X, aid 0x%02X)' % (flag,aid),'parser')
		return self.parse(left,negotiated)

	def merge_attributes (self):
		as2path = self[Attribute.CODE.AS_PATH]
		as4path = self[Attribute.CODE.AS4_PATH]
		self.remove(Attribute.CODE.AS_PATH)
		self.remove(Attribute.CODE.AS4_PATH)

		# this key is unique as index length is a two header, plus a number of ASN of size 2 or 4
		# so adding the: make the length odd and unique
		key = "%s:%s" % (as2path.index, as4path.index)

		# found a cache copy
		cached = Attribute.cache.get(Attribute.CODE.AS_PATH,{}).get(key,None)
		if cached:
			self.add(cached,key)
			return

		# as_seq = []
		# as_set = []

		len2 = len(as2path.as_seq)
		len4 = len(as4path.as_seq)

		# RFC 4893 section 4.2.3
		if len2 < len4:
			as_seq = as2path.as_seq
		else:
			as_seq = as2path.as_seq[:-len4]
			as_seq.extend(as4path.as_seq)

		len2 = len(as2path.as_set)
		len4 = len(as4path.as_set)

		if len2 < len4:
			as_set = as4path.as_set
		else:
			as_set = as2path.as_set[:-len4]
			as_set.extend(as4path.as_set)

		aspath = ASPath(as_seq,as_set)
		self.add(aspath,key)

	def __hash__ (self):
		# FIXME: two routes with distinct nh but other attributes equal
		# will hash to the same value until repr represents the nh (??)
		return hash(repr(self))

	def __eq__ (self, other):
		return self.sameValuesAs(other)

	# BaGPipe code ..

	# test that sets of attributes exactly match
	# can't rely on __eq__ for this, because __eq__ relies on Attribute.__eq__ which does not look at attributes values

	def sameValuesAs (self, other):
		# we sort based on packed values since the items do not
		# necessarily implement __cmp__
		def pack_(x):
			return x.pack()

		try:
			for key in set(six.iterkeys(self)).union(set(six.iterkeys(other))):
				if (key == Attribute.CODE.MP_REACH_NLRI or key == Attribute.CODE.MP_UNREACH_NLRI):
					continue

				sval = self[key]
				oval = other[key]

				# In the case where the attribute is Communities or
				# extended communities, we want to compare values independently of their order
				if isinstance(sval, Communities):
					if not isinstance(oval, Communities):
						return False

					sval = sorted(sval,key=pack_)
					oval = sorted(oval,key=pack_)

				if sval != oval:
					return False
			return True
		except KeyError:
			return False



# encoding: utf-8
"""
node.py

Created by Evelio Vila on 2016-11-26. eveliovila@gmail.com
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from struct import pack
from struct import unpack

from exabgp.protocol.ip import IP
from exabgp.protocol.iso import ISO
from exabgp.bgp.message.notification import Notify
from exabgp.util import ordinal

#           +--------------------+-------------------+----------+
#           | Sub-TLV Code Point | Description       |   Length |
#           +--------------------+-------------------+----------+
#           |        512         | Autonomous System |        4 |
#           |        513         | BGP-LS Identifier |        4 |
#           |        514         | OSPF Area-ID      |        4 |
#           |        515         | IGP Router-ID     | Variable |
#           +--------------------+-------------------+----------+
#            https://tools.ietf.org/html/rfc7752#section-3.2.1.4
# ================================================================== NODE-DESC-SUB-TLVs

NODE_TLVS = {
	512:	'autonomous-system',
	513:	'bgp-ls-id',
	514:	'ospf-area-id',
	515:	'igp-rid',
}


# TODO
# 3.2.1.5.  Multi-Topology ID

class NodeDescriptor (object):
	def __init__ (self, node_id, dtype, psn=None, dr_id=None, packed=None):
		self.node_id = node_id
		self.dtype = dtype
		self.psn = psn
		self.dr_id = dr_id
		self._packed = packed

	@classmethod
	def unpack (cls, data, igp):
		dtype, dlength = unpack('!HH',data[0:4])
		if dtype not in NODE_TLVS.keys():
			raise Exception("Unknown Node Descriptor Sub-TLV")
		# OSPF Area-ID
		if dtype == 514:
			return cls(
				node_id=IP.unpack(data[4: 4 + dlength]),
				dtype=dtype,
				packed=data[:4+dlength]
			), data[4 + dlength:]
		# IGP Router-ID: The TLV size in combination with the protocol
		# identifier enables the decoder to determine the type
		# of the node: sec 3.2.1.4.
		elif dtype == 515:
			# OSPFv{2,3} non-pseudonode
			if (igp == 3 or igp == 6) and dlength == 4:
				r_id = IP.unpack(data[4: 4 + 4])
				return cls(
					node_id=r_id,
					dtype=dtype,
					packed=data[:4+dlength]
				), data[4 + 4:]
			# OSPFv{2,3} LAN pseudonode
			if (igp == 3 or igp == 6) and dlength == 8:
				r_id = IP.unpack(data[4: 4 + 4])
				dr_id = IP.unpack(data[8: 4 + 8])
				return cls(
					node_id=r_id,
					dtype=dtype,
					psn=None,
					dr_id=dr_id,
					packed=data[:4+dlength]
				), data[4 + 8:]
			# IS-IS non-pseudonode
			if (igp == 1 or igp == 2) and dlength == 6:
				return cls(
					node_id=ISO.unpack_sysid(data[4: 4 + 6]),
					dtype=dtype,
					packed=data[:4+dlength]
				), data[4 + 6:]
			# IS-IS LAN pseudonode = ISO Node-ID + PSN
			# Unpack ISO address
			if (igp == 1 or igp == 2) and dlength == 7:
				iso_node = ISO.unpack_sysid(data[4: 4 + 6])
				psn = unpack('!B', data[4 + 6: 4 + 7])[0]
				return cls(
					node_id=iso_node,
					dtype=dtype,
					psn=psn,
					packed=data[:4+dlength]
				), data[4 + 7:]
		elif dtype == 512 and dlength == 4:
			# ASN
			return cls(
				node_id=unpack('!L', data[4: 4 + dlength])[0],
				dtype=dtype,
				packed=data[:4+dlength]
			), data[4 + 4:]
		elif dtype == 513 and dlength == 4:
			# BGP-LS
			return cls(
				node_id=unpack('!L',data[4: 4 + dlength])[0],
				dtype=dtype,
				packed=data[:4+dlength]
			), data[4 + 4:]
		else:
			raise Notify(3,5,'could not decode Local Node descriptor')

	def json (self,compact=None):
		ospf = None
		designated = None
		psn = None
		router_id = None
		asn = None
		bgpls_id = None
		if self.dtype == 514:
			ospf = '"ospf-area-id": "%s"' % self.node_id
		if self.dr_id is not None:
			designated = '"designated-router-id": "%s"' % self.dr_id
		if self.psn is not None:
			psn = '"psn": "%s"' % self.psn
		if self.dtype == 515:
			router_id = '"router-id": "%s"' % self.node_id
		if self.dtype == 512:
			asn = '"autonomous-system": %d' % self.node_id
		if self.dtype == 513:
			bgpls_id = '"bgp-ls-identifier": "%d"' % self.node_id
		content = ', '.join(d for d in [ospf,designated,psn,router_id,asn,bgpls_id] if d)
		return content

	def __eq__ (self, other):
		return isinstance(other, NodeDescriptor) and self.node_id == other.node_id

	def __neq__ (self, other):
		return self.node_id != other.node_id

	def __lt__ (self, other):
		raise RuntimeError('Not implemented')

	def __le__ (self, other):
		raise RuntimeError('Not implemented')

	def __gt__ (self, other):
		raise RuntimeError('Not implemented')

	def __ge__ (self, other):
		raise RuntimeError('Not implemented')

	def __str__ (self):
		return self.json()

	def __repr__ (self):
		return self.__str__()

	def __len__ (self):
		return len(self._packed)

	def __hash__ (self):
		return hash(str(self))

	def pack (self):
		return self._packed



#!/usr/bin/env python
# encoding: utf-8
"""
protocol.py

Created by Thomas Mangin on 2009-08-27.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

import unittest

from exabgp.bgp.message.open.asn import ASN

from exabgp.bgp.message import Open
from exabgp.bgp.message.open import Capabilities
# from exabgp.bgp.message import Notification
from exabgp.bgp.message import KeepAlive
from exabgp.bgp.message import Update
from exabgp.bgp.message.update import Attributes

from exabgp.reactor.protocol import Protocol
from exabgp.bgp.neighbor import Neighbor

from exabgp.vendoring import six
from exabgp.vendoring.six.moves import StringIO

from exabgp.configuration.environment import environment
environment.setup('')


class Network (StringIO):
	def pending (self):
		return True


# route1 = Update([],[to_NLRI('10.0.0.1','32')],Attributes())
# route1.next_hop = '10.0.0.254'
#
#
# route2 = Update([],[to_NLRI('10.0.1.1','32')],Attributes())
# route2.next_hop = '10.0.0.254'
#
#
# route3 = Update([],[to_NLRI('10.0.2.1','32')],Attributes())
# route3.next_hop = '10.0.0.254'
#
#
# routes = [route1,route2,route3]
# routes.sort()


class TestProtocol (unittest.TestCase):

	def setUp (self):
		self.neighbor = Neighbor()
		self.neighbor.local_as = ASN(65000)
		self.neighbor.peer_as = ASN(65000)
		# self.neighbor.peer_address = InetIP('1.2.3.4')
		# self.neighbor.local_address = InetIP('5.6.7.8')

	def test_1_selfparse_open (self):
		# ds = Open(4,65000,'1.2.3.4',Capabilities().default(),30)
		#
		# txt = ds.message()
		# network = Network(txt)
		# print [hex(ord(c)) for c in txt]
		# bgp = Protocol(self.neighbor,network)
		# bgp.follow = False

		# o = bgp.read_open('127.0.0.1')
		# self.assertEqual(o.version,4)
		# self.assertEqual(o.asn,65000)
		# self.assertEqual(o.hold_time,30)
		# self.assertEqual(str(o.router_id),'1.2.3.4')
		pass

	def test_2_selfparse_KeepAlive (self):
		# ds = KeepAlive()
		#
		# txt = ds.message()
		# network = Network(txt)
		# bgp = Protocol(self.neighbor,network)

		# message = bgp.read_message()
		# self.assertEqual(message.TYPE,KeepAlive.TYPE)
		pass

	def test_3_parse_update (self):
		# txt = ''.join([chr(c) for c in [0x0, 0x0, 0x0, 0x1c, 0x40, 0x1, 0x1, 0x2, 0x40, 0x2, 0x0, 0x40, 0x3, 0x4, 0xc0, 0x0, 0x2, 0xfe, 0x80, 0x4, 0x4, 0x0, 0x0, 0x0, 0x0, 0x40, 0x5, 0x4, 0x0, 0x0, 0x1, 0x23, 0x20, 0x52, 0xdb, 0x0, 0x7, 0x20, 0x52, 0xdb, 0x0, 0x45, 0x20, 0x52, 0xdb, 0x0, 0x47]])
		# updates = new_Update(txt)
		#
		# routes = [str(route) for route in updates.added()]
		# routes.sort()
		# self.assertEqual(routes[0],'82.219.0.69/32 next-hop 192.0.2.254')
		# self.assertEqual(routes[1],'82.219.0.7/32 next-hop 192.0.2.254')
		# self.assertEqual(routes[2],'82.219.0.71/32 next-hop 192.0.2.254')
		pass

	def test_4_parse_update (self):
		# txt = ''.join([chr(c) for c in [0x0, 0x0, 0x0, 0x12, 0x40, 0x1, 0x1, 0x0, 0x40, 0x2, 0x4, 0x2, 0x1, 0x78, 0x14, 0x40, 0x3, 0x4, 0x52, 0xdb, 0x2, 0xb5, 0x0]])
		# updates = new_Update(txt)
		# self.assertEqual(str(updates.added()[0]),'0.0.0.0/0 next-hop 82.219.2.181')
		pass

	def test_6_holdtime (self):
		# class MyPeer(Network):
		# 	_data = StringIO(Open(4,65000,'1.2.3.4',Capabilities().default(),90).message())
		# 	def read (self, l):
		# 		return self._data.read(l)
		#
		# network = MyPeer('')
		#
		# bgp = Protocol(self.neighbor,network)
		# bgp.follow = False
		#
		# before = bgp.neighbor.hold_time
		# bgp.new_open()
		# bgp.read_open('127.0.0.1')
		# after = bgp.neighbor.hold_time
		#
		# self.assertEqual(after,min(before,90))
		pass

	def test_7_message (self):
		# txt = ''.join([chr(_) for _ in [0x0, 0x0, 0x0, 0x30, 0x40, 0x1, 0x1, 0x0, 0x50, 0x2, 0x0, 0x4, 0x2, 0x1, 0xff, 0xfe, 0x80, 0x4, 0x4, 0x0, 0x0, 0x0, 0x0, 0x80, 0xe, 0x1a, 0x0, 0x2, 0x1, 0x10, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x20, 0x12, 0x34, 0x56, 0x78]])
		# updates = new_Update(txt)
		# print updates
		# self.assertEqual(str(updates.added()[0]),'1234:5678::/32 next-hop ::')
		pass

	def test_7_ipv6 (self):
		# txt = ''.join([chr(_) for _ in [0x0, 0x0, 0x0, 0x25, 0x40, 0x1, 0x1, 0x0, 0x40, 0x2, 0x4, 0x2, 0x1, 0xfd, 0xe8, 0xc0, 0x8, 0x8, 0x78, 0x14, 0x0, 0x0, 0x78, 0x14, 0x78, 0x14, 0x40, 0xf, 0xc, 0x0, 0x2, 0x1, 0x40, 0x2a, 0x2, 0xb, 0x80, 0x0, 0x0, 0x0, 0x1]])
		# updates = new_Update(txt)
		pass

if __name__ == '__main__':
	unittest.main()



# encoding: utf-8
"""
l2vpn/__init__.py

Created by Thomas Mangin on 2015-06-04.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from exabgp.configuration.l2vpn.vpls import ParseVPLS

from exabgp.bgp.message.update.nlri import VPLS
from exabgp.bgp.message.update.attribute import Attributes
from exabgp.rib.change import Change

from exabgp.configuration.announce import ParseAnnounce


class ParseL2VPN (ParseVPLS):
	syntax = \
		'vpls %s;\n' % ' '.join(ParseVPLS.definition)

	action = dict(ParseVPLS.action)

	name = 'L2VPN'

	def __init__ (self, tokeniser, scope, error, logger):
		ParseVPLS.__init__(self,tokeniser,scope,error,logger)

	def clear (self):
		return True

	def pre (self):
		return True

	def post (self):
		routes = self.scope.pop_routes()
		if routes:
			self.scope.extend('routes',routes)
		return True


@ParseL2VPN.register('vpls','append-route')
def vpls (tokeniser):
	change = Change(
		VPLS(None,None,None,None,None),
		Attributes()
	)

	while True:
		command = tokeniser()

		if not command:
			break

		action = ParseVPLS.action[command]

		if 'nlri-set' in action:
			change.nlri.assign(ParseVPLS.assign[command],ParseL2VPN.known[command](tokeniser))
		elif 'attribute-add' in action:
			change.attributes.add(ParseL2VPN.known[command](tokeniser))
		elif action == 'nexthop-and-attribute':
			nexthop,attribute = ParseVPLS.known[command](tokeniser)
			change.nlri.nexthop = nexthop
			change.attributes.add(attribute)
		else:
			raise ValueError('vpls: unknown command "%s"' % command)

	return [change,]



# encoding: utf-8
"""
sr/srgb.py

Created by Evelio Vila 2017-02-16
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
"""
import json
from struct import pack

from exabgp.util import concat_bytes
from exabgp.vendoring.bitstring import BitArray
from exabgp.bgp.message.notification import Notify
from exabgp.bgp.message.update.attribute.sr.prefixsid import PrefixSid

# 0                   1                   2                   3
# 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |     Type      |          Length               |    Flags      |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |     Flags     |
# +-+-+-+-+-+-+-+-+
#
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |         SRGB 1 (6 octets)                                     |
# |                               +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |                               |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |         SRGB n (6 octets)                                     |
# |                               +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |                               |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# 3.3.  Originator SRGB TLV


@PrefixSid.register()
class SrGb(object):
	TLV = 3
	# Length is the total length of the value portion of the TLV: 2 +
	# multiple of 6.
	LENGTH = -1

	def __init__ (self, srgbs, packed=None):
		self.srgbs = srgbs
		self.packed = self.pack()

	def __repr__ (self):
		items = []
		for base, srange in self.srgbs:
			items.append("( {},{} )".format(base, srange))
		return '[ {} ]'.format(', '.join(items))

	def pack (self):
		payload = pack('!H',0)  # flags
		for b,r in self.srgbs:
			payload = concat_bytes(
				payload,
				pack("!L",b)[1:],
				pack("!L",r)[1:]
			)
		return concat_bytes(
			pack('!B', self.TLV),
			pack('!H', len(payload)),
			payload
		)

	@classmethod
	def unpack (cls,data,length):
		srgbs = []
		# Flags: 16 bits of flags.  None is defined by this document.  The
		# flag field MUST be clear on transmission and MUST be ignored at
		# reception.
		data = data[2:]
		# SRGB: 3 octets of base followed by 3 octets of range.  Note that
		# the SRGB field MAY appear multiple times.  If the SRGB field
		# appears multiple times, the SRGB consists of multiple ranges.
		while data:
			b = BitArray(bytes=data[:3])
			base = b.unpack('uintbe:24')[0]
			b = BitArray(bytes=data[3:6])
			srange = b.unpack('uintbe:24')[0]
			srgbs.append((base,srange))
			data = data[6:]
		return cls(srgbs=srgbs)

	def json (self,compact=None):
		return '"sr-srgbs": {}'.format((json.dumps(self.srgbs)))



# encoding: utf-8
"""
segment.py

Created by Thomas Mangin on 2014-06-27.
Copyright (c) 2014-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from exabgp.protocol.ip import IP

from exabgp.util import character
from exabgp.util import ordinal
from exabgp.util import concat_bytes

from exabgp.bgp.message.update.nlri.qualifier import RouteDistinguisher
from exabgp.bgp.message.update.nlri.qualifier import ESI

from exabgp.bgp.message.update.nlri.evpn.nlri import EVPN

from exabgp.bgp.message.notification import Notify

# +---------------------------------------+
# |      RD   (8 octets)                  |
# +---------------------------------------+
# |Ethernet Segment Identifier (10 octets)|
# +---------------------------------------+
# |  IP Address Length (1 octet)          |
# +---------------------------------------+
# |   Originating Router's IP Addr        |
# |          (4 or 16 octets)             |
# +---------------------------------------+

# ===================================================================== EVPNNLRI


@EVPN.register
class EthernetSegment (EVPN):
	CODE = 4
	NAME = "Ethernet Segment"
	SHORT_NAME = "Segment"

	def __init__ (self, rd, esi, ip, packed=None,nexthop=None,action=None,addpath=None):
		EVPN.__init__(self,action,addpath)
		self.nexthop = nexthop
		self.rd = rd
		self.esi = esi
		self.ip = ip
		self._pack(packed)

	def __eq__ (self, other):
		return \
			isinstance(other, EthernetSegment) and \
			self.CODE == other.CODE and \
			self.rd == other.rd and \
			self.ip == other.ip
		# esi and label must not be part of the comparaison

	def __ne__ (self, other):
		return not self.__eq__(other)

	def __str__ (self):
		return "%s:%s:%s:%s" % (
			self._prefix(),
			self.rd._str(),
			self.esi,
			self.ip if self.ip else ""
		)

	def __hash__ (self):
		# esi and label MUST *NOT* be part of the hash
		return hash((self.rd,self.ip))

	def _pack (self, packed=None):
		if self._packed:
			return self._packed

		if packed:
			self._packed = packed
			return packed

		self._packed = concat_bytes(
			self.rd.pack(),
			self.esi.pack(),
			character(len(self.ip)*8 if self.ip else 0),
			self.ip.pack() if self.ip else b''
		)
		return self._packed

	@classmethod
	def unpack (cls, data):
		rd = RouteDistinguisher.unpack(data[:8])
		esi = ESI.unpack(data[8:18])
		iplen = ordinal(data[18])

		if iplen not in (32,128):
			raise Notify(3,5,"IP length field is given as %d in current Segment, expecting 32 (IPv4) or 128 (IPv6) bits" % iplen)

		ip = IP.unpack(data[19:19+(iplen//8)])

		return cls(rd,esi,ip,data)

	def json (self, compact=None):
		content = ' "code": %d, ' % self.CODE
		content += '"parsed": true, '
		content += '"raw": "%s", ' % self._raw()
		content += '"name": "%s", ' % self.NAME
		content += '%s, ' % self.rd.json()
		content += self.esi.json()
		content += ', "ip": "%s"' % str(self.ip)
		return '{%s }' % content



# encoding: utf-8
"""
store.py

Created by Thomas Mangin on 2009-11-05.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from exabgp.rib.cache import Cache


class IncomingRIB (Cache):
	def __init__ (self, cache, families):
		Cache.__init__(self,cache,families)

	# back to square one, all the routes are removed
	def clear (self):
		self.clear_cache()

	def reset (self):
		pass



#!/usr/bin/env python
# encoding: utf-8
"""
flow.py

Created by Thomas Mangin on 2010-01-14.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

import socket
import unittest

from exabgp.util import character
from exabgp.util import ordinal
from exabgp.util import concat_strs

from exabgp.bgp.message.update.nlri import Flow
from exabgp.bgp.message.update.nlri.flow import Flow4Source
from exabgp.bgp.message.update.nlri.flow import Flow4Destination
from exabgp.bgp.message.update.nlri.flow import FlowAnyPort

from exabgp.bgp.message.update.nlri.flow import NumericOperator
# from exabgp.bgp.message.update.attribute.community import *

from exabgp.protocol.ip import IPv4

from exabgp.util.test import data_from_body

from exabgp.configuration.environment import environment
environment.setup('')


class TestFlow (unittest.TestCase):

	def setUp (self):
		pass

	def test_rule (self):
		components = {
			'destination': Flow4Destination(IPv4.pton("192.0.2.0"),24),
			'source':      Flow4Source(IPv4.pton("10.1.2.0"),24),
			'anyport_1':   FlowAnyPort(NumericOperator.EQ,25),
		}
		messages = {
			'destination': [0x01, 0x18, 0xc0, 0x00, 0x02],
			'source':      [0x02, 0x18, 0x0a, 0x01, 0x02],
			'anyport_1':   [0x04, 0x01, 0x19],
		}

		for key in ['destination','source','anyport_1']:
			component = components[key].pack()
			message   = data_from_body(messages[key])
			# if component != message:
			# 	self.fail('content mismatch\n%s\n%s' % (['0x%02X' % ord(_) for _ in component],['0x%02X' % ord(_) for _ in message]))

	def test_rule_and (self):
		components = {
			'destination': Flow4Destination(IPv4.pton("192.0.2.0"),24),
			'source':      Flow4Source(IPv4.pton("10.1.2.0"),24),
			'anyport_1':   FlowAnyPort(NumericOperator.EQ | NumericOperator.GT,25),
			'anyport_2':   FlowAnyPort(NumericOperator.EQ | NumericOperator.LT,80),
		}
		messages = {
			'destination': [0x01, 0x18, 0xc0, 0x00, 0x02],
			'source':      [0x02, 0x18, 0x0a, 0x01, 0x02],
			'anyport_1':   [0x04, 0x43, 0x19],
			'anyport_2':   [0x04, 0x85, 0x50],
		}

		flow = Flow()
		message = b""
		for key in ['destination','source','anyport_1','anyport_2']:
			flow.add(components[key])
			message += data_from_body(messages[key])
		message = character(len(message)) + message
		# flow.add(to_FlowAction(65000,False,False))
		flow.pack()
		# print [hex(ord(_)) for _ in flow]

	def test_nlri (self):
		components = {
			'destination': Flow4Destination(IPv4.pton("192.0.2.0"),24),
			'source':      Flow4Source(IPv4.pton("10.1.2.0"),24),
			'anyport_1':   FlowAnyPort(NumericOperator.EQ | NumericOperator.GT,25),
			'anyport_2':   FlowAnyPort(NumericOperator.EQ | NumericOperator.LT,80),
		}
		messages = {
			'destination': [0x01, 0x18, 0xc0, 0x00, 0x02],
			'source':      [0x02, 0x18, 0x0a, 0x01, 0x02],
			'anyport_1':   [0x04, 0x43, 0x19],
			'anyport_2':   [0x85, 0x50],
		}

		flow = Flow()
		message = b""
		for key in ['destination','source','anyport_1','anyport_2']:
			flow.add(components[key])
			message += data_from_body(messages[key])
		message = character(len(message)) + message
		# policy.add(to_FlowAction(65000,False,False))
		flow = flow.pack()
		if message[0] != flow[0]:
			self.fail('size mismatch %s %s\n' % (ordinal(flow[0]),ordinal(message[0])))
		if len(flow) != ordinal(flow[0]) + 1:
			self.fail('invalid size for message')
		# if message[1:] != flow[1:]:
		# 	self.fail('content mismatch\n%s\n%s' % (['0x%02X' % ord(_) for _ in flow],['0x%02X' % ord(_) for _ in message]))

	def test_compare (self):
		components = {
			'destination': Flow4Destination(IPv4.pton("192.0.2.0"),24),
			'source':      Flow4Source(IPv4.pton("10.1.2.0"),24),
			'anyport_1':   FlowAnyPort(NumericOperator.EQ | NumericOperator.GT,25),
			'anyport_2':   FlowAnyPort(NumericOperator.EQ | NumericOperator.LT,80),
			'anyport_3':   FlowAnyPort(NumericOperator.EQ,80),
		}

		flow1 = Flow()
		for key in ['destination','source','anyport_1','anyport_2']:
			flow1.add(components[key])

		flow2 = Flow()
		for key in ['destination','source','anyport_3']:
			flow2.add(components[key])

		if flow1 != flow1:
			self.fail('the flows are the same')

		if flow1 == flow2:
			self.fail('the flows are not the same')

if __name__ == '__main__':
	unittest.main()



"""
nlri.py

Created by Thomas Morin on 2014-06-23.
Copyright (c) 2014-2017 Orange. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from struct import pack

from exabgp.protocol.ip import NoNextHop

from exabgp.protocol.family import AFI
from exabgp.protocol.family import SAFI

from exabgp.util import ordinal
from exabgp.bgp.message import OUT

from exabgp.bgp.message.update.nlri import NLRI

# https://tools.ietf.org/html/rfc7432

# +-----------------------------------+
# |    Route Type (1 octet)           |
# +-----------------------------------+
# |     Length (1 octet)              |
# +-----------------------------------+
# | Route Type specific (variable)    |
# +-----------------------------------+

# ========================================================================= EVPN


@NLRI.register(AFI.l2vpn,SAFI.evpn)
class EVPN (NLRI):
	registered_evpn = dict()

	# NEED to be defined in the subclasses
	CODE = -1
	NAME = 'Unknown'
	SHORT_NAME = 'unknown'

	def __init__ (self, action=OUT.UNSET, addpath=None):
		NLRI.__init__(self, AFI.l2vpn, SAFI.evpn, action)
		self._packed = b''

	def __hash__ (self):
		return hash("%s:%s:%s:%s" % (self.afi,self.safi,self.CODE,self._packed))

	def __len__ (self):
		return len(self._packed) + 2

	def __eq__ (self, other):
		return \
			NLRI.__eq__(self,other) and \
			self.CODE == other.CODE

	def __str__ (self):
		return "evpn:%s:%s" % (self.registered_evpn.get(self.CODE,self).SHORT_NAME.lower(),'0x' + ''.join('%02x' % ordinal(_) for _ in self._packed))

	def __repr__ (self):
		return str(self)

	def feedback (self, action):
		# if self.nexthop is None and action == OUT.ANNOUNCE:
		# 	return 'evpn nlri next-hop is missing'
		return ''

	def _prefix (self):
		return "evpn:%s:" % (self.registered_evpn.get(self.CODE,self).SHORT_NAME.lower())

	def pack_nlri (self, negotiated=None):
		# XXX: addpath not supported yet
		return pack('!BB',self.CODE,len(self._packed)) + self._packed

	@classmethod
	def register (cls, klass):
		if klass.CODE in cls.registered_evpn:
			raise RuntimeError('only one EVPN registration allowed')
		cls.registered_evpn[klass.CODE] = klass
		return klass

	@classmethod
	def unpack_nlri (cls, afi, safi, bgp, action, addpath):
		code = ordinal(bgp[0])
		length = ordinal(bgp[1])

		if code in cls.registered_evpn:
			klass = cls.registered_evpn[code].unpack(bgp[2:length+2])
		else:
			klass = GenericEVPN(code,bgp[2:length+2])
		klass.CODE = code
		klass.action = action
		klass.addpath = addpath

		return klass,bgp[length+2:]

	def _raw (self):
		return ''.join('%02X' % ordinal(_) for _ in self.pack_nlri())


class GenericEVPN (EVPN):
	def __init__ (self,code,packed):
		EVPN.__init__(self)
		self.CODE = code
		self._pack(packed)

	def _pack (self,packed=None):
		if self._packed:
			return self._packed

		if packed:
			self._packed = packed
			return packed

	def json (self, compact=None):
		return '{ "code": %d, "parsed": false, "raw": "%s" }' % (self.CODE,self._raw())



# encoding: utf-8
"""
bgp.py

Created by Thomas Mangin on 2009-08-30.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

import os
import sys
import stat
import platform
import syslog
import string

from exabgp.util.dns import warn
from exabgp.logger import Logger

from exabgp.version import version
# import before the fork to improve copy on write memory savings
from exabgp.reactor.loop import Reactor

from exabgp.vendoring import docopt
from exabgp.vendoring import lsprofcalltree

from exabgp.configuration.usage import usage

from exabgp.debug import setup_report
setup_report()


def is_bgp (s):
	return all(c in string.hexdigits or c == ':' for c in s)


def __exit (memory, code):
	if memory:
		from exabgp.vendoring import objgraph
		sys.stdout.write('memory utilisation\n\n')
		sys.stdout.write(objgraph.show_most_common_types(limit=20))
		sys.stdout.write('\n\n\n')
		sys.stdout.write('generating memory utilisation graph\n\n')
		sys.stdout.write()
		obj = objgraph.by_type('Reactor')
		objgraph.show_backrefs([obj], max_depth=10)
	sys.exit(code)


def named_pipe (root, pipename='exabgp'):
	locations = [
		'/run/exabgp/',
		'/run/%d/' % os.getuid(),
		'/run/',
		'/var/run/exabgp/',
		'/var/run/%d/' % os.getuid(),
		'/var/run/',
		root + '/run/exabgp/',
		root + '/run/%d/' % os.getuid(),
		root + '/run/',
		root + '/var/run/exabgp/',
		root + '/var/run/%d/' % os.getuid(),
		root + '/var/run/',
	]
	for location in locations:
		cli_in = location + pipename + '.in'
		cli_out = location + pipename + '.out'

		try:
			if not stat.S_ISFIFO(os.stat(cli_in).st_mode):
				continue
			if not stat.S_ISFIFO(os.stat(cli_out).st_mode):
				continue
		except KeyboardInterrupt:
			raise
		except Exception:
			continue
		os.environ['exabgp_cli_pipe'] = location
		return [location]
	return locations


def root_folder (options,locations):
	if options['--root']:
		return os.path.realpath(os.path.normpath(options['--root'])).rstrip('/')

	argv = os.path.realpath(os.path.normpath(os.path.join(os.getcwd(),sys.argv[0])))

	for location in locations:
		if argv.endswith(location):
			return argv[:-len(location)]
	return ''


def get_envfile (options, etc):
	envfile = 'exabgp.env' if not options["--env"] else options["--env"]
	if not envfile.startswith('/'):
		envfile = '%s/%s' % (etc, envfile)
	return envfile


def get_env (envfile):
	from exabgp.configuration.setup import environment

	try:
		return environment.setup(envfile)
	except environment.Error as exc:
		sys.stdout.write(usage)
		sys.stdout.flush()
		print('\nconfiguration issue,', str(exc))
		sys.exit(1)


def main ():
	major = int(sys.version[0])
	minor = int(sys.version[2])

	if major <= 2 and minor < 5:
		sys.stdout.write('This program can not work (is not tested) with your python version (< 2.5)\n')
		sys.stdout.flush()
		sys.exit(1)

	cli_named_pipe = os.environ.get('exabgp_cli_pipe','')
	if cli_named_pipe:
		from exabgp.application.control import main as control
		control(cli_named_pipe)
		sys.exit(0)

	options = docopt.docopt(usage, help=False)

	if options["--run"]:
		sys.argv = sys.argv[sys.argv.index('--run')+1:]
		if sys.argv[0] == 'healthcheck':
			from exabgp.application import run_healthcheck
			run_healthcheck()
		elif sys.argv[0] == 'cli':
			from exabgp.application import run_cli
			run_cli()
		else:
			sys.stdout.write(usage)
			sys.stdout.flush()
			sys.exit(0)
		return

	root = root_folder(options,['/bin/exabgp','/sbin/exabgp','/lib/exabgp/application/bgp.py','/lib/exabgp/application/control.py'])
	prefix = '' if root == '/usr' else root
	etc = prefix + '/etc/exabgp'

	os.environ['EXABGP_ETC'] = etc  # This is not most pretty

	if options["--version"]:
		sys.stdout.write('ExaBGP : %s\n' % version)
		sys.stdout.write('Python : %s\n' % sys.version.replace('\n',' '))
		sys.stdout.write('Uname  : %s\n' % ' '.join(platform.uname()[:5]))
		sys.stdout.write('Root   : %s\n' % root)
		sys.stdout.flush()
		sys.exit(0)

	envfile = get_envfile(options,etc)
	env = get_env(envfile)

	# Must be done before setting the logger as it modify its behaviour
	if options["--debug"]:
		env.log.all = True
		env.log.level = syslog.LOG_DEBUG

	logger = Logger()

	from exabgp.configuration.setup import environment

	if options["--decode"]:
		decode = ''.join(options["--decode"]).replace(':','').replace(' ','')
		if not is_bgp(decode):
			sys.stdout.write(usage)
			sys.stdout.write('Environment values are:\n%s\n\n' % '\n'.join(' - %s' % _ for _ in environment.default()))
			sys.stdout.write('The BGP message must be an hexadecimal string.\n\n')
			sys.stdout.write('All colons or spaces are ignored, for example:\n\n')
			sys.stdout.write('  --decode 001E0200000007900F0003000101\n')
			sys.stdout.write('  --decode 001E:02:0000:0007:900F:0003:0001:01\n')
			sys.stdout.write('  --decode FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF001E0200000007900F0003000101\n')
			sys.stdout.write('  --decode FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF:001E:02:0000:0007:900F:0003:0001:01\n')
			sys.stdout.write('  --decode \'FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF 001E02 00000007900F0003000101\n\'')
			sys.stdout.flush()
			sys.exit(1)
	else:
		decode = ''

	duration = options["--signal"]
	if duration and duration.isdigit():
		pid = os.fork()
		if pid:
			import time
			import signal
			try:
				time.sleep(int(duration))
				os.kill(pid,signal.SIGUSR1)
			except KeyboardInterrupt:
				pass
			try:
				pid,code = os.wait()
				sys.exit(code)
			except KeyboardInterrupt:
				try:
					pid,code = os.wait()
					sys.exit(code)
				except Exception:
					sys.exit(0)

	if options["--help"]:
		sys.stdout.write(usage)
		sys.stdout.write('Environment values are:\n' + '\n'.join(' - %s' % _ for _ in environment.default()))
		sys.stdout.flush()
		sys.exit(0)

	if options["--decode"]:
		env.log.parser = True
		env.debug.route = decode
		env.tcp.bind = ''

	if options["--profile"]:
		env.profile.enable = True
		if options["--profile"].lower() in ['1','true']:
			env.profile.file = True
		elif options["--profile"].lower() in ['0','false']:
			env.profile.file = False
		else:
			env.profile.file = options["--profile"]

	if envfile and not os.path.isfile(envfile):
		comment = 'environment file missing\ngenerate it using "exabgp --fi > %s"' % envfile
	else:
		comment = ''

	if options["--full-ini"] or options["--fi"]:
		for line in environment.iter_ini():
			sys.stdout.write('%s\n' % line)
			sys.stdout.flush()
		sys.exit(0)

	if options["--full-env"] or options["--fe"]:
		print()
		for line in environment.iter_env():
			sys.stdout.write('%s\n' % line)
			sys.stdout.flush()
		sys.exit(0)

	if options["--diff-ini"] or options["--di"]:
		for line in environment.iter_ini(True):
			sys.stdout.write('%s\n' % line)
			sys.stdout.flush()
		sys.exit(0)

	if options["--diff-env"] or options["--de"]:
		for line in environment.iter_env(True):
			sys.stdout.write('%s\n' % line)
			sys.stdout.flush()
		sys.exit(0)

	if options["--once"]:
		env.tcp.once = True

	if options["--pdb"]:
		# The following may fail on old version of python (but is required for debug.py)
		os.environ['PDB'] = 'true'
		env.debug.pdb = True

	if options["--test"]:
		env.debug.selfcheck = True
		env.log.parser = True

	if options["--memory"]:
		env.debug.memory = True

	configurations = []
	# check the file only once that we have parsed all the command line options and allowed them to run
	if options["<configuration>"]:
		for f in options["<configuration>"]:
			normalised = os.path.realpath(os.path.normpath(f))
			if os.path.isfile(normalised):
				configurations.append(normalised)
				continue
			if f.startswith('etc/exabgp'):
				normalised = os.path.join(etc,f[11:])
				if os.path.isfile(normalised):
					configurations.append(normalised)
					continue

			logger.debug('one of the arguments passed as configuration is not a file (%s)' % f,'configuration')
			sys.exit(1)

	else:
		sys.stdout.write(usage)
		sys.stdout.write('Environment values are:\n%s\n\n' % '\n'.join(' - %s' % _ for _ in environment.default()))
		sys.stdout.write('no configuration file provided')
		sys.stdout.flush()
		sys.exit(1)

	from exabgp.bgp.message.update.attribute import Attribute
	Attribute.caching = env.cache.attributes

	if env.debug.rotate or len(configurations) == 1:
		run(env,comment,configurations,root,options["--validate"])

	if not (env.log.destination in ('syslog','stdout','stderr') or env.log.destination.startswith('host:')):
		logger.error('can not log to files when running multiple configuration (as we fork)','configuration')
		sys.exit(1)

	try:
		# run each configuration in its own process
		pids = []
		for configuration in configurations:
			pid = os.fork()
			if pid == 0:
				run(env,comment,[configuration],root,options["--validate"],os.getpid())
			else:
				pids.append(pid)

		# If we get a ^C / SIGTERM, ignore just continue waiting for our child process
		import signal
		signal.signal(signal.SIGINT, signal.SIG_IGN)

		# wait for the forked processes
		for pid in pids:
			os.waitpid(pid,0)
	except OSError as exc:
		logger.critical('can not fork, errno %d : %s' % (exc.errno,exc.strerror),'reactor')
		sys.exit(1)


def run (env, comment, configurations, root, validate, pid=0):
	logger = Logger()

	logger.notice('Thank you for using ExaBGP','welcome')
	logger.notice('%s' % version,'version')
	logger.notice('%s' % sys.version.replace('\n',' '),'interpreter')
	logger.notice('%s' % ' '.join(platform.uname()[:5]),'os')
	logger.notice('%s' % root,'installation')

	if comment:
		logger.notice(comment,'advice')

	warning = warn()
	if warning:
		logger.warning(warning,'advice')

	if env.api.cli:
		pipename = 'exabgp' if env.api.pipename is None else env.api.pipename
		pipes = named_pipe(root, pipename)
		if len(pipes) != 1:
			env.api.cli = False
			logger.error('could not find the named pipes (%s.in and %s.out) required for the cli' % (pipename,pipename),'cli')
			logger.error('we scanned the following folders (the number is your PID):','cli')
			for location in pipes:
				logger.error(' - %s' % location,'cli control')
			logger.error('please make them in one of the folder with the following commands:','cli control')
			logger.error('> mkfifo %s/run/%s.{in,out}' % (os.getcwd(),pipename),'cli control')
			logger.error('> chmod 600 %s/run/%s.{in,out}' % (os.getcwd(),pipename) ,'cli control')
			if os.getuid() != 0:
				logger.error('> chown %d:%d %s/run/%s.{in,out}' % (os.getuid(),os.getgid(),os.getcwd(),pipename),'cli control')
		else:
			pipe = pipes[0]
			os.environ['exabgp_cli_pipe'] = pipe
			os.environ['exabgp_api_pipename'] = pipename

			logger.info('named pipes for the cli are:','cli control')
			logger.info('to send commands  %s%s.in' % (pipe,pipename),'cli control')
			logger.info('to read responses %s%s.out' % (pipe,pipename),'cli control')

	if not env.profile.enable:
		exit_code = Reactor(configurations).run(validate,root)
		__exit(env.debug.memory, exit_code)

	try:
		import cProfile as profile
	except ImportError:
		import profile

	if env.profile.file == 'stdout':
		profiled = 'Reactor(%s).run(%s,"%s")' % (str(configurations),str(validate),str(root))
		exit_code = profile.run(profiled)
		__exit(env.debug.memory, exit_code)

	if pid:
		profile_name = "%s-pid-%d" % (env.profile.file,pid)
	else:
		profile_name = env.profile.file

	notice = ''
	if os.path.isdir(profile_name):
		notice = 'profile can not use this filename as output, it is not a directory (%s)' % profile_name
	if os.path.exists(profile_name):
		notice = 'profile can not use this filename as output, it already exists (%s)' % profile_name

	if not notice:
		cwd = os.getcwd()
		logger.debug('profiling ....','reactor')
		profiler = profile.Profile()
		profiler.enable()
		try:
			exit_code = Reactor(configurations).run(validate, root)
		except Exception:
			exit_code = Reactor.Exit.unknown
			raise
		finally:
			profiler.disable()
			kprofile = lsprofcalltree.KCacheGrind(profiler)
			try:
				destination = profile_name if profile_name.startswith('/') else os.path.join(cwd,profile_name)
				with open(destination, 'w+') as write:
					kprofile.output(write)
			except IOError:
				notice = 'could not save profiling in formation at: ' + destination
				logger.debug("-"*len(notice),'reactor')
				logger.debug(notice,'reactor')
				logger.debug("-"*len(notice),'reactor')
			__exit(env.debug.memory,exit_code)
	else:
		logger.debug("-"*len(notice),'reactor')
		logger.debug(notice,'reactor')
		logger.debug("-"*len(notice),'reactor')
		Reactor(configurations).run(validate,root)
		__exit(env.debug.memory,1)


if __name__ == '__main__':
	main()



# encoding: utf-8
"""
Copyright (c) 2016 Evelio Vila <eveliovila@gmail.com>
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

import json
import binascii
import itertools
from struct import unpack

from exabgp.util import concat_strs
from exabgp.vendoring.bitstring import BitArray
from exabgp.bgp.message.notification import Notify
from exabgp.bgp.message.update.attribute.attribute import Attribute


@Attribute.register()
class LINKSTATE(Attribute):
	ID = Attribute.CODE.BGP_LS
	FLAG = Attribute.Flag.OPTIONAL
	TLV = -1

	# Registered subclasses we know how to decode
	registered_lsids = dict()

	# what this implementation knows as LS attributes
	node_lsids = []
	link_lsids = []
	prefix_lsids = []

	def __init__(self, ls_attrs):
		self.ls_attrs = ls_attrs

	@classmethod
	def register(cls,lsid=None,flag=None):
		def register_lsid (klass):
			scode = klass.TLV if lsid is None else lsid
			if scode in cls.registered_lsids:
				raise RuntimeError('only one class can be registered per BGP link state attribute type')
			cls.registered_lsids[scode] = klass
			return klass
		return register_lsid

	@classmethod
	def registered(cls, lsid, flag=None):
		return lsid in cls.registered_lsids

	@classmethod
	def unpack(cls, data, negotiated):
		ls_attrs = []
		while data:
			scode, length = unpack('!HH',data[:4])
			if scode in cls.registered_lsids:
				klass = cls.registered_lsids[scode].unpack(data[4:length+4],length)
			else:
				klass = GenericLSID(scode,data[4:length+4])
			klass.TLV = scode
			ls_attrs.append(klass)
			data = data[length+4:]

		return cls(ls_attrs=ls_attrs)

	def json (self,compact=None):
		content = ', '.join(d.json() for d in self.ls_attrs)
		return '{ %s }' % (content)

	def __str__(self):
		return ', '.join(str(d) for d in self.ls_attrs)


class GenericLSID(object):
	TLV = 99999

	def __init__ (self, code, rep):
		self.rep = rep
		self.code = code

	def __repr__ (self):
		return "Attribute with code [ %s ] not implemented" % (self.code)

	@classmethod
	def unpack (cls,scode,data):
		length = len(data)
		info = binascii.b2a_uu(data[:length])
		return cls(code=scode,rep=info)

	def json (self,compact=None):
		return '"attribute-not-implemented": "%s"' % (self.code)


class LsGenericFlags(object):

#	draft-ietf-isis-segment-routing-extensions Prefix-SID Sub-TLV
	ISIS_SR_FLAGS = ['R', 'N', 'P', 'E', 'V', 'L', 'RSV', 'RSV']
#	RFC 7794 IPv4/IPv6 Extended Reachability Attribute Flags
	ISIS_SR_ATTR_FLAGS = ['X', 'R', 'N', 'RSV', 'RSV', 'RSV', 'RSV', 'RSV']
#	draft-ietf-isis-segment-routing-extensions - Adj-SID IS-IS Flags
	ISIS_SR_ADJ_FLAGS = ['F', 'B', 'V', 'L', 'S', 'P', 'RSV', 'RSV']
# 	isis-segment-routing-extensions 3.1. SR-Capabilities Sub-TLV
	ISIS_SR_CAP_FLAGS = ['I', 'V', 'RSV', 'RSV', 'RSV', 'RSV', 'RSV', 'RSV']
#	RFC 7752 3.3.1.1. Node Flag Bits TLV
	LS_NODE_FLAGS = ['O', 'T', 'E', 'B', 'R', 'V', 'RSV', 'RSV']
#	RFC 7752 3.3.3.1. IGP Flags TLV
	LS_IGP_FLAGS = ['D', 'N', 'L', 'P', 'RSV', 'RSV', 'RSV', 'RSV']
#	RFC 7752 3.3.2.2.  MPLS Protocol Mask TLV
	LS_MPLS_MASK = ['LDP', 'RSVP-TE', 'RSV', 'RSV', 'RSV', 'RSV', 'RSV', 'RSV']
#	RFC 5307 1.2.
	LS_PROTECTION_MASK = ['ExtraTrafic', 'Unprotected', 'Shared', 'Dedicated 1:1',
							'Dedicated 1+1', 'Enhanced', 'RSV', 'RSV']

	def __init__(self,flags):
		self.flags = flags

	@classmethod
	def unpack(cls,data,pattern):
		pad = pattern.count('RSV')
		repeat = len(pattern) - pad
		flag_array = binascii.b2a_hex(data)
		hex_rep = hex(int(flag_array, 16))
		bit_array = BitArray(hex_rep)
		valid_flags = [concat_strs(''.join(item), ''.join(itertools.repeat('0',pad)))
				for item in itertools.product('01', repeat=repeat)]
		valid_flags.append('0000')
		if bit_array.bin in valid_flags:
			flags = dict(zip(pattern, bit_array.bin))
			flags = {k:int(v) for k, v in flags.items()}
		else:
			raise Notify(3,5, "Invalid SR flags mask")
		return cls(flags=flags)

	def json (self,compact=None):
    		return json.dumps(self.flags)

	def __repr__ (self):
		return str(self.flags)



# encoding: utf-8
"""
maxbw.py

Created by Evelio Vila on 2016-12-01.
Copyright (c) 2014-2017 Exa Networks. All rights reserved.
"""

from struct import unpack

from exabgp.bgp.message.notification import Notify
from exabgp.bgp.message.update.attribute.bgpls.linkstate import LINKSTATE

#  This sub-TLV contains the maximum bandwidth that can be used on this
#   link in this direction (from the system originating the LSP to its
#   neighbors).
#    https://tools.ietf.org/html/rfc5305#section-3.4
#
#  Units are in Bytes not Bits.
#  ----------------------------

@LINKSTATE.register()
class MaxBw(object):
	TLV = 1089

	def __init__ (self, maxbw):
		self.maxbw = maxbw

	def __repr__ (self):
		return "Maximum link bandwidth: %s" % (self.maxbw)

	@classmethod
	def unpack (cls,data,length):
		if length != 4:
			raise Notify(3,5, "Incorrect maximum link bw metric")
		else:
			maxbw = unpack('!f',data)[0]
			return cls(maxbw=maxbw)

	def json (self,compact=None):
		return '"maximum-link-bandwidth": %s' % self.maxbw



# encoding: utf-8
"""
tokeniser.py

Created by Thomas Mangin on 2015-06-05.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from exabgp.configuration.core.format import tokens
from exabgp.protocol.family import AFI
from collections import deque
from exabgp.vendoring import six


class Tokeniser (object):

	class Iterator (object):
		fname = ''  # This is ok to have a unique value as API parser do not use files

		def __init__ (self):
			self.next = deque()
			self.tokens = []
			self.generator = iter([])

		def replenish (self, content):
			self.next.clear()
			self.tokens = content
			self.generator = iter(content)
			return self

		def clear (self):
			self.replenish([])

		def __call__ (self):
			if self.next:
				return self.next.popleft()

			try:
				return six.next(self.generator)
			except StopIteration:
				return ''

		def peek (self):
			try:
				peaked = six.next(self.generator)
				self.next.append(peaked)
				return peaked
			except StopIteration:
				return ''

	@staticmethod
	def _off ():
		return iter([])

	def __init__ (self, scope, error, logger):
		self.scope = scope
		self.error = error
		self.logger = logger
		self.finished = False
		self.number = 0
		self.line = []
		self.iterate = Tokeniser.Iterator()
		self.end = ''
		self.index_column = 0
		self.index_line = 0
		self.fname = ''
		self.type = 'unset'
		self.afi = AFI.undefined

		self._tokens = Tokeniser._off
		self._next = None
		self._data = None

	def clear (self):
		self.finished = False
		self.number = 0
		self.line = []
		self.iterate.clear()
		self.end = ''
		self.index_column = 0
		self.index_line = 0
		self.fname = ''
		self.type = 'unset'
		if self._data:
			self._set(self._data)

	def params (self):
		if len(self.line) <= 2:
			return ''
		if self.end in ('{','}',';'):
			return "'%s'" % "' '".join(self.line[1:-1])
		return "'%s'" % "' '".join(self.line[1:])

	def _tokenise (self,iterator):
		for parsed in tokens(iterator):
			words = [word for y,x,word in parsed]
			self.line = ''.join(words)
			# ignore # lines
			# set Location information
			yield words

	def _set (self, function):
		try:
			self._tokens = function
			self._next = six.next(self._tokens)
		except IOError as exc:
			error = str(exc)
			if error.count(']'):
				self.error.set(error.split(']')[1].strip())
			else:
				self.error.set(error)
			self._tokens = Tokeniser._off
			self._next = []
			return self.error.set('issue setting the configuration parser')
		except StopIteration:
			self._tokens = Tokeniser._off
			self._next = []
			return self.error.set('issue setting the configuration parser, no data')
		return True

	def set_file (self, data):
		def _source (fname):
			with open(fname,'r') as fileobject:
				def formated ():
					line = ''
					for current in fileobject:
						self.index_line += 1
						current = current.rstrip()
						if current.endswith('\\'):
							line += current
							continue
						elif line:
							yield line + current
							line = ''
						else:
							yield current
					if line:
						yield line

				for _ in self._tokenise(formated()):
					yield _
		self.type = 'file'
		self.Iterator.fname = data
		return self._set(_source(data))

	def set_text (self, data):
		def _source (data):
			for _ in self._tokenise(data.split('\n')):
				yield _
		self.type = 'text'
		return self._set(_source(data))

	def set_api (self, line):
		return self._set(self._tokenise(iter([line])))

	def __call__ (self):
		self.number += 1
		try:
			self.line, self._next = self._next, six.next(self._tokens)
			self.end = self.line[-1]
		except StopIteration:
			if not self.finished:
				self.finished = True
				self.line, self._next = self._next, []
				self.end = self.line[-1]
			else:
				self.line = []
				self.end = ''

		# should we raise a Location if called with no more data ?
		self.iterate.replenish(self.line[:-1])

		return self.line



#!/usr/bin/env python

import sys
import time

def write (data):
	sys.stdout.write(data + '\n')
	sys.stdout.flush()

def main ():
	msg = 'announce attribute next-hop 1.2.3.4 med 100 as-path [ 100 101 102 103 104 105 106 107 108 109 110 ] nlri %s'
	write(msg % ' '.join('%d.0.0.0/8' % ip for ip in range(0,224)))
	write(msg % ' '.join('10.%d.0.0/16' % ip for ip in range(0,256)))

	time.sleep(2)

	write('withdraw attribute next-hop 1.2.3.4 med 100 as-path [ 100 101 102 103 104 105 106 107 108 109 110 ] nlri 0.0.0.0/8 1.0.0.0/8')

	time.sleep(10000)

if __name__ == '__main__':
	try:
		main()
	except KeyboardInterrupt:
		pass



# encoding: utf-8
"""
vpls.py

Created by Thomas Mangin on 2015-06-05.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from exabgp.configuration.core import Section

from exabgp.configuration.static.parser import attribute
from exabgp.configuration.static.parser import origin
from exabgp.configuration.static.parser import med
from exabgp.configuration.static.parser import as_path
from exabgp.configuration.static.parser import local_preference
from exabgp.configuration.static.parser import atomic_aggregate
from exabgp.configuration.static.parser import aggregator
from exabgp.configuration.static.parser import originator_id
from exabgp.configuration.static.parser import cluster_list
from exabgp.configuration.static.parser import community
from exabgp.configuration.static.parser import extended_community
from exabgp.configuration.static.parser import name as named
from exabgp.configuration.static.parser import split
from exabgp.configuration.static.parser import watchdog
from exabgp.configuration.static.parser import withdraw

from exabgp.configuration.static.mpls import route_distinguisher

from exabgp.configuration.l2vpn.parser import vpls
from exabgp.configuration.l2vpn.parser import vpls_endpoint
from exabgp.configuration.l2vpn.parser import vpls_size
from exabgp.configuration.l2vpn.parser import vpls_offset
from exabgp.configuration.l2vpn.parser import vpls_base
from exabgp.configuration.l2vpn.parser import next_hop


class ParseVPLS (Section):
	definition = [
		'endpoint <vpls endpoint id; integer>',
		'base <label base; integer>',
		'offset <block offet; interger>',
		'size <block size; integer>',

		'next-hop <ip>',
		'med <16 bits number>',
		'route-distinguisher|rd <ipv4>:<port>|<16bits number>:<32bits number>|<32bits number>:<16bits number>',
		'origin IGP|EGP|INCOMPLETE',
		'as-path [ <asn>.. ]',
		'local-preference <16 bits number>',
		'atomic-aggregate',
		'community <16 bits number>',
		'extended-community target:<16 bits number>:<ipv4 formated number>',
		'originator-id <ipv4>',
		'cluster-list <ipv4>',
		'label <15 bits number>',
		'attribute [ generic attribute format ]'
		'name <mnemonic>',
		'split /<mask>',
		'watchdog <watchdog-name>',
		'withdraw',
	]

	syntax = \
		'vpls {\n  %s\n}' % ' ;\n  '.join(definition)

	known = {
		'rd':                 route_distinguisher,
		'attribute':          attribute,
		'next-hop':           next_hop,
		'origin':             origin,
		'med':                med,
		'as-path':            as_path,
		'local-preference':   local_preference,
		'atomic-aggregate':   atomic_aggregate,
		'aggregator':         aggregator,
		'originator-id':      originator_id,
		'cluster-list':       cluster_list,
		'community':          community,
		'extended-community': extended_community,
		'name':               named,
		'split':              split,
		'watchdog':           watchdog,
		'withdraw':           withdraw,
		'endpoint':           vpls_endpoint,
		'offset':             vpls_offset,
		'size':               vpls_size,
		'base':               vpls_base,
	}

	action = {
		'attribute':           'attribute-add',
		'origin':              'attribute-add',
		'med':                 'attribute-add',
		'as-path':             'attribute-add',
		'local-preference':    'attribute-add',
		'atomic-aggregate':    'attribute-add',
		'aggregator':          'attribute-add',
		'originator-id':       'attribute-add',
		'cluster-list':        'attribute-add',
		'community':           'attribute-add',
		'extended-community':  'attribute-add',
		'name':                'attribute-add',
		'split':               'attribute-add',
		'watchdog':            'attribute-add',
		'withdraw':            'attribute-add',
		'next-hop':            'nlri-set',
		'route-distinguisher': 'nlri-set',
		'rd':                  'nlri-set',
		'endpoint':            'nlri-set',
		'offset':              'nlri-set',
		'size':                'nlri-set',
		'base':                'nlri-set',
	}

	assign = {
		'next-hop':            'nexthop',
		'rd':                  'rd',
		'route-distinguisher': 'rd',
		'endpoint':            'endpoint',
		'offset':              'offset',
		'size':                'size',
		'base':                'base',
	}

	name = 'l2vpn/vpls'

	def __init__ (self, tokeniser, scope, error, logger):
		Section.__init__(self,tokeniser,scope,error,logger)

	def clear (self):
		pass

	def pre (self):
		self.scope.append_route(vpls(self.tokeniser.iterate))
		return True

	def post (self):
		if not self._check():
			return False
		return True

	def _check (self):
		feedback = self.scope.get_route().feedback()
		if feedback:
			return self.error.set(feedback)
		return True



# encoding: utf-8
"""
notification.py

Created by Thomas Mangin on 2009-11-05.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

import string

from exabgp.util import character
from exabgp.util import ordinal
from exabgp.util import concat_bytes
from exabgp.util import str_ascii
from exabgp.util import bytes_ascii
from exabgp.util import hexbytes
from exabgp.util import hexstring

from exabgp.bgp.message.message import Message


# ================================================================== Notification
# A Notification received from our peer.
# RFC 4271 Section 4.5

# 0                   1                   2                   3
# 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# | Error code    | Error subcode |   Data (variable)             |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

@Message.register
class Notification (Message):
	ID = Message.CODE.NOTIFICATION
	TYPE = character(Message.CODE.NOTIFICATION)

	_str_code = {
		1: "Message header error",
		2: "OPEN message error",
		3: "UPDATE message error",
		4: "Hold timer expired",
		5: "State machine error",
		6: "Cease"
	}

	_str_subcode = {
		(1,0): "Unspecific",
		(1,1): "Connection Not Synchronized",
		(1,2): "Bad Message Length",
		(1,3): "Bad Message Type",

		(2,0): "Unspecific",
		(2,1): "Unsupported Version Number",
		(2,2): "Bad Peer AS",
		(2,3): "Bad BGP Identifier",
		(2,4): "Unsupported Optional Parameter",
		(2,5): "Authentication Notification (Deprecated)",
		(2,6): "Unacceptable Hold Time",
		# RFC 5492 - https://tools.ietf.org/html/rfc5492
		(2,7): "Unsupported Capability",

		# draft-ietf-idr-bgp-multisession-06
		(2,8):  "Grouping Conflict",
		(2,9):  "Grouping Required",
		(2,10): "Capability Value Mismatch",

		(3,0): "Unspecific",
		(3,1): "Malformed Attribute List",
		(3,2): "Unrecognized Well-known Attribute",
		(3,3): "Missing Well-known Attribute",
		(3,4): "Attribute Flags Error",
		(3,5): "Attribute Length Error",
		(3,6): "Invalid ORIGIN Attribute",
		(3,7): "AS Routing Loop",
		(3,8): "Invalid NEXT_HOP Attribute",
		(3,9): "Optional Attribute Error",
		(3,10): "Invalid Network Field",
		(3,11): "Malformed AS_PATH",

		(4,0): "Unspecific",

		(5,0): "Unspecific",
		# RFC 6608 - https://tools.ietf.org/html/rfc6608
		(5,1): "Receive Unexpected Message in OpenSent State",
		(5,2): "Receive Unexpected Message in OpenConfirm State",
		(5,3): "Receive Unexpected Message in Established State",

		(6,0): "Unspecific",
		# RFC 4486 - https://tools.ietf.org/html/rfc4486
		(6,1): "Maximum Number of Prefixes Reached",
		(6,2): "Administrative Shutdown",  # augmented with draft-ietf-idr-shutdown
		(6,3): "Peer De-configured",
		(6,4): "Administrative Reset",
		(6,5): "Connection Rejected",
		(6,6): "Other Configuration Change",
		(6,7): "Connection Collision Resolution",
		(6,8): "Out of Resources",

		# draft-keyur-bgp-enhanced-route-refresh-00
		(7,1): "Invalid Message Length",
		(7,2): "Malformed Message Subtype",
	}

	def __init__ (self, code, subcode, data=b'', parse_data=True):
		self.code = code
		self.subcode = subcode

		if not parse_data:
			self.data = data
			return

		if not (code, subcode) in [(6, 2), (6, 4)]:
			self.data = data if not len([_ for _ in str(data) if _ not in string.printable]) else hexbytes(data)
			return

		if len(data) == 0:
			# shutdown without shutdown communication (the old fashioned way)
			self.data = b''
			return

		# draft-ietf-idr-shutdown or the peer was using 6,2 with data

		shutdown_length  = ordinal(data[0])
		data = data[1:]

		if shutdown_length == 0:
			self.data = b"empty Shutdown Communication."
			# move offset past length field
			return

		if len(data) < shutdown_length:
			self.data = b"invalid Shutdown Communication (buffer underrun) length : %i [%s]" % (shutdown_length, hexstring(data))
			return

		if shutdown_length > 128:
			self.data = b"invalid Shutdown Communication (too large) length : %i [%s]" % (shutdown_length, hexstring(data))
			return

		try:
			self.data = b'Shutdown Communication: "%s"' % \
				data[:shutdown_length].decode('utf-8').replace('\r',' ').replace('\n',' ')
		except UnicodeDecodeError:
			self.data = b"invalid Shutdown Communication (invalid UTF-8) length : %i [%s]" % (shutdown_length, hexstring(data))
			return

		trailer = data[shutdown_length:]
		if trailer:
			self.data += b", trailing data: " + hexstring(trailer)

	def __str__ (self):
		return "%s / %s%s" % (
			self._str_code.get(self.code,'unknown error'),
			self._str_subcode.get((self.code,self.subcode),'unknow reason'),
			'%s' % (' / %s' % str_ascii(self.data) if self.data else '')
		)

	@classmethod
	def unpack_message (cls, data, negotiated=None):
		return cls(ordinal(data[0]),ordinal(data[1]),data[2:])


# =================================================================== Notify
# A Notification we need to inform our peer of.

@Message.notify
class Notify (Notification):
	def __init__ (self, code, subcode, data=None):
		if data is None:
			data = self._str_subcode.get((code,subcode),'unknown notification type')
		if (code, subcode) in [(6, 2), (6, 4)]:
			data = chr(len(data)) + data
		Notification.__init__(self,code,subcode,bytes_ascii(data),False)

	def message (self,negotiated=None):
		return self._message(
			character(self.code) +
			character(self.subcode) +
			self.data
		)



# encoding: utf-8
"""
parse_process.py

Created by Thomas Mangin on 2015-06-18.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

import os
import stat


def encoder (tokeniser):
	value = tokeniser()

	if value not in ('text','json'):
		raise ValueError('"%s" is an invalid option' % value)
	return value


def _make_path (prg):
	parts = prg.split('/')

	env = os.environ.get('EXABGP_ETC','')
	if env:
		options = [
			os.path.join(env.rstrip('/'),os.path.join(*parts[2:])),
			'/etc/exabgp'
		]
	else:
		options = []
		options.append('/etc/exabgp')
		pwd = os.environ.get('PWD','').split('/')
		if pwd:
			# without abspath the path is not / prefixed !
			if pwd[-1] in ('etc','sbin'):
				options.append(os.path.abspath(os.path.join(os.path.join(*pwd[:-1]),os.path.join(*parts))))
			if 'etc' not in pwd:
				options.append(os.path.abspath(os.path.join(os.path.join(*pwd),os.path.join(*parts))))
	return options


def run (tokeniser):
	prg = tokeniser()

	if prg[0] != '/':
		if prg.startswith('etc/exabgp'):
			options = _make_path(prg)
		else:
			options = [
				os.path.abspath(os.path.join('/etc/exabgp',prg)),
				os.path.abspath(os.path.join(os.path.dirname(tokeniser.fname),prg)),
			]
			options.extend((os.path.abspath(os.path.join(p,prg)) for p in os.getenv('PATH').split(':')))
		for option in options:
			if os.path.exists(option):
				prg = option

	if not os.path.exists(prg):
		raise ValueError('can not locate the the program "%s"' % prg)

	# race conditions are possible, those are sanity checks not security ones ...
	s = os.stat(prg)

	if stat.S_ISDIR(s.st_mode):
		raise ValueError('can not execute directories "%s"' % prg)

	if s.st_mode & stat.S_ISUID:
		raise ValueError('refusing to run setuid programs "%s"' % prg)

	check = stat.S_IXOTH
	if s.st_uid == os.getuid():
		check |= stat.S_IXUSR
	if s.st_gid == os.getgid():
		check |= stat.S_IXGRP

	if not check & s.st_mode:
		raise ValueError('exabgp will not be able to run this program "%s"' % prg)

	return [prg] + [_ for _ in tokeniser.generator]



# encoding: utf-8
"""
asn.py

Created by Thomas Mangin on 2010-01-15.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from struct import pack
from struct import unpack
import sys

from exabgp.protocol.resource import Resource

if sys.version_info > (3,):
	long = int

# =================================================================== ASN


class ASN (Resource):
	MAX = pow(2,16)-1

	def asn4 (self):
		return self > self.MAX

	def pack (self, negotiated=None):
		asn4 = negotiated if negotiated is not None else self.asn4()
		return pack('!L' if asn4 else '!H',self)

	@classmethod
	def unpack (cls, data, klass=None):
		kls = cls if klass is None else klass
		value = unpack('!L' if len(data) == 4 else '!H',data)[0]
		return kls(value)

	def __len__ (self):
		return 4 if self.asn4() else 2

	def extract (self):
		return [pack('!L',self)]

	def trans (self):
		if self.asn4():
			return AS_TRANS
		return self

	def __repr__ (self):
		return '%ld' % long(self)

	def __str__ (self):
		return '%ld' % long(self)

	@classmethod
	def from_string (cls, value):
		return cls(long(value))


AS_TRANS = ASN(23456)



# encoding: utf-8
"""
graceful.py

Created by Thomas Mangin on 2012-07-17.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from struct import pack
from struct import unpack

from exabgp.vendoring import six

from exabgp.util import character
from exabgp.util import ordinal
from exabgp.util import concat_bytes_i
from exabgp.util import concat_bytes

from exabgp.protocol.family import AFI
from exabgp.protocol.family import SAFI
from exabgp.bgp.message.open.capability.capability import Capability

# =========================================================== Graceful (Restart)
# RFC 4727 - https://tools.ietf.org/html/rfc4727


@Capability.register()
class Graceful (Capability,dict):
	MAX = 0xFFFF
	ID = Capability.CODE.GRACEFUL_RESTART

	TIME_MASK     = 0x0FFF
	FLAG_MASK     = 0xF000

	# 0x8 is binary 1000
	RESTART_STATE = 0x08
	FORWARDING_STATE = 0x80

	def set (self, restart_flag, restart_time, protos):
		self.restart_flag = restart_flag
		self.restart_time = restart_time & Graceful.TIME_MASK
		for afi,safi,family_flag in protos:
			self[(afi,safi)] = family_flag & Graceful.FORWARDING_STATE
		return self

	def extract (self):
		restart  = pack('!H',((self.restart_flag << 12) | (self.restart_time & Graceful.TIME_MASK)))
		families = [(afi.pack(),safi.pack(),character(self[(afi,safi)])) for (afi,safi) in self.keys()]
		sfamilies = concat_bytes_i(concat_bytes(pafi,psafi,family) for (pafi,psafi,family) in families)
		return [concat_bytes(restart,sfamilies)]

	def __str__ (self):
		families = [(str(afi),str(safi),hex(self[(afi,safi)])) for (afi,safi) in self.keys()]
		sfamilies = ' '.join(["%s/%s=%s" % (afi,safi,family) for (afi,safi,family) in families])
		return "Graceful Restart Flags %s Time %d %s" % (hex(self.restart_flag),self.restart_time,sfamilies)

	def json (self):
		d = {
			'name':'"graceful restart"',
			'time':self.restart_time,
			'address-family-flags':'{ %s}' % (
				', '.join('"%s/%s": [%s ] ' % (
					'%s' % afi, safi, ' "restart"' if family & 0x80 else '') for afi, safi, family in [
						(str(a), str(s), self[(a,s)]) for (a,s) in self.keys()
					]
				)
			),
			'restart-flags':'[%s] ' % (' "forwarding" ' if self.restart_flag & 0x8 else ' ')
		}

		return '{ %s}' % ', '.join('"%s": %s' % (k,v) for k,v in six.iteritems(d))

	def families (self):
		return self.keys()

	@staticmethod
	def unpack_capability (instance, data, capability=None):  # pylint: disable=W0613
		# XXX: FIXME: should raise if instance was already setup
		restart = unpack('!H',data[:2])[0]
		restart_flag = restart >> 12
		restart_time = restart & Graceful.TIME_MASK
		data = data[2:]
		families = []
		while data:
			afi = AFI.unpack(data[:2])
			safi = SAFI.unpack(data[2])
			flag_family = ordinal(data[3])
			families.append((afi,safi,flag_family))
			data = data[4:]
		return instance.set(restart_flag,restart_time,families)
